from os import path, makedirs
import numpy as np
from htmd.adaptive.adaptive import AdaptiveBase
from protocolinterface import val
from htmd.projections.tica import TICA
from htmd.projections.metric import Metric
from htmd.clustering.regular import RegCluster
from htmd.adaptive.util import getParentSimIdxFrame, updatingMean
from sklearn.cluster import MiniBatchKMeans
import logging
logger = logging.getLogger(__name__)


class AdaptiveBandit(AdaptiveBase):
    """

    Parameters
    ----------
    app : :class:`SimQueue <jobqueues.simqueue.SimQueue>` object, default=None
        A SimQueue class object used to retrieve and submit simulations
    project : str, default='adaptive'
        The name of the project
    nmin : int, default=1
        Minimum number of running simulations
    nmax : int, default=1
        Maximum number of running simulations
    nepochs : int, default=1000
        Stop adaptive once we have reached this number of epochs
    nframes : int, default=0
        Stop adaptive once we have simulated this number of aggregate simulation frames.
    inputpath : str, default='input'
        The directory used to store input folders
    generatorspath : str, default='generators'
        The directory containing the generators
    dryrun : boolean, default=False
        A dry run means that the adaptive will retrieve and generate a new epoch but not submit the simulations
    updateperiod : float, default=0
        When set to a value other than 0, the adaptive will run synchronously every `updateperiod` seconds
    coorname : str, default='input.coor'
        Name of the file containing the starting coordinates for the new simulations
    lock : bool, default=False
        Lock the folder while adaptive is ongoing
    datapath : str, default='data'
        The directory in which the completed simulations are stored
    filter : bool, default=True
        Enable or disable filtering of trajectories.
    filtersel : str, default='not water'
        Filtering atom selection
    filteredpath : str, default='filtered'
        The directory in which the filtered simulations will be stored
    projection : :class:`Projection <moleculekit.projections.projection.Projection>` object, default=None
        A Projection class object or a list of objects which will be used to project the simulation data before
        constructing a Markov model
    goalfunction : function, default=None
        This function will be used to convert the goal-projected simulation data to a ranking whichcan be used for the
        directed component of FAST.
    reward_method : str, default='max'
        The reward method
    skip : int, default=1
        Allows skipping of simulation frames to reduce data. i.e. skip=3 will only keep every third frame
    lag : int, default=1
        The lagtime used to create the Markov model. Units are in frames.
    exploration : float, default=0.5
        Exploration is the coefficient used in UCB algorithm to weight the exploration value
    temperature : int, default=300
        Temperature used to compute the free energy
    ticalag : int, default=20
        Lagtime to use for TICA in frames. When using `skip` remember to change this accordinly.
    ticadim : int, default=3
        Number of TICA dimensions to use. When set to 0 it disables TICA
    clustmethod : :class:`ClusterMixin <sklearn.base.ClusterMixin>` class, default=<class 'sklearn.cluster.k_means_.MiniBatchKMeans'>
        Clustering algorithm used to cluster the contacts or distances
    macronum : int, default=8
        The number of macrostates to produce
    save : bool, default=False
        Save the model generated
    save_qval : bool, default=False
        Save the Q(a) and N values for every epoch
    actionspace : str, default='metric'
        The action space
    recluster : bool, default=False
        If to recluster the action space.
    reclusterMethod : , default=<class 'sklearn.cluster.k_means_.MiniBatchKMeans'>
        Clustering method for reclustering.
    goal_init : float, default=0.3
        The proportional ratio of goal initialization compared to max frames set by nframes
    actionpool : int, default=0
        The number of top scoring actions used to randomly select respawning simulations
    """
    def __init__(self):
        from sklearn.base import ClusterMixin
        from moleculekit.projections.projection import Projection
        super().__init__()
        self._arg('datapath', 'str', 'The directory in which the completed simulations are stored', 'data', val.String())
        self._arg('filter', 'bool', 'Enable or disable filtering of trajectories.', True, val.Boolean())
        self._arg('filtersel', 'str', 'Filtering atom selection', 'not water', val.String())
        self._arg('filteredpath', 'str', 'The directory in which the filtered simulations will be stored', 'filtered', val.String())
        self._arg('projection', ':class:`Projection <moleculekit.projections.projection.Projection>` object',
                  'A Projection class object or a list of objects which will be used to project the simulation '
                   'data before constructing a Markov model', None, val.Object(Projection), nargs='+')
        self._arg('goalfunction', 'function',
                  'This function will be used to convert the goal-projected simulation data to a ranking which'
                  'can be used for the directed component of FAST.', None, val.Function(), nargs='any')
        self._arg('reward_method', 'str', 'The reward method', 'mean', val.String()) # Default should be mean
        self._arg('skip', 'int', 'Allows skipping of simulation frames to reduce data. i.e. skip=3 will only keep every third frame', 1, val.Number(int, 'POS'))
        self._arg('lag', 'int', 'The lagtime used to create the Markov model. Units are in frames.', 1, val.Number(int, 'POS'))
        self._arg('exploration', 'float', 'Exploration is the coefficient used in UCB algorithm to weight the exploration value', 0.01, val.Number(float, 'OPOS')) # Default changed to 0.01
        self._arg('temperature', 'int', 'Temperature used to compute the free energy', 300, val.Number(int, 'POS'))
        self._arg('ticalag', 'int', 'Lagtime to use for TICA in frames. When using `skip` remember to change this accordinly.', 20, val.Number(int, '0POS'))
        self._arg('ticadim', 'int', 'Number of TICA dimensions to use. When set to 0 it disables TICA', 3, val.Number(int, '0POS'))
        self._arg('clustmethod', ':class:`ClusterMixin <sklearn.base.ClusterMixin>` class', 'Clustering algorithm used to cluster the contacts or distances', MiniBatchKMeans, val.Class(ClusterMixin))
        self._arg('macronum', 'int', 'The number of macrostates to produce', 8, val.Number(int, 'POS'))
        self._arg('save', 'bool', 'Save the model generated', False, val.Boolean())
        self._arg('save_qval', 'bool', 'Save the Q(a) and N values for every epoch', False, val.Boolean())
        self._arg('actionspace', 'str', 'The action space', 'tica', val.String())
        self._arg('recluster', 'bool', 'If to recluster the action space.', False, val.Boolean())
        self._arg('reclusterMethod', '', 'Clustering method for reclustering.', MiniBatchKMeans)
        self._arg('goal_init', 'float', 'The proportional ratio of goal initialization compared to max frames set by nframes', 0.3, val.Number(float, 'POS'))
        self._arg('actionpool', 'int', 'The number of top scoring actions used to randomly select respawning simulations', 0, val.Number(int, 'OPOS'))

    def _algorithm(self):
        from htmd.kinetics import Kinetics
        sims = self._getSimlist()
        metr = Metric(sims, skip=self.skip)
        metr.set(self.projection)

        data = metr.project()
        data.dropTraj()  # Drop before TICA to avoid broken trajectories

        if self.goalfunction is not None:
            goaldata = self._getGoalData(data.simlist)
            if len(data.simlist) != len(goaldata.simlist):
                raise RuntimeError('The goal function was not able to project all trajectories that the MSM projection could. Check for possible errors in the goal function.')
            goaldataconcat = np.concatenate(goaldata.dat)
            if self.save:
                makedirs('saveddata', exist_ok=True)
                goaldata.save(path.join('saveddata', 'e{}_goaldata.dat'.format(self._getEpoch())))

        # tica = TICA(metr, int(max(2, np.ceil(self.ticalag))))  # gianni: without project it was tooooo slow
        if self.ticadim > 0:
            ticalag = int(np.ceil(max(2, min(np.min(data.trajLengths) / 2, self.ticalag))))  # 1 < ticalag < (trajLen / 2)
            tica = TICA(data, ticalag)
            datatica = tica.project(self.ticadim)
            if not self._checkNFrames(datatica): return False
            self._createMSM(datatica)
        else:
            if not self._checkNFrames(data): return False
            self._createMSM(data)

        confstatdist = self.conformationStationaryDistribution(self._model)
        if self.actionspace == 'metric':
            if not data.K:
                data.cluster(self.clustmethod(n_clusters=self._numClusters(data.numFrames)))
            data_q = data.copy()
        elif self.actionspace == 'goal':
            data_q = goaldata.copy()
        elif self.actionspace == 'tica':
            data_q = datatica.copy()
        elif self.actionspace == 'ticapcca':
            data_q = datatica.copy()
            for traj in data_q.trajectories:
                traj.cluster = self._model.macro_ofcluster[traj.cluster]
            data_q.K = self._model.macronum

        if self.recluster:
            print('Reclustering with {}'.format(self.reclusterMethod))
            data_q.cluster(self.reclusterMethod)
        
        numstates = data_q.K
        print('Numstates: {}'.format(numstates))
        currepoch = self._getEpoch()
        q_values = np.zeros(numstates, dtype=np.float32)
        n_values = np.zeros(numstates, dtype=np.int32)

        if self.goalfunction is not None:
            ## For every cluster in data_q, get the max score and initialize
            qstconcat = np.concatenate(data_q.St)
            statemaxes = np.zeros(numstates)
            np.maximum.at(statemaxes, qstconcat, np.squeeze(goaldataconcat))

            goalenergies = -Kinetics._kB * self.temperature * np.log(1-statemaxes)
            q_values = goalenergies
            n_values += int((self.nframes / self._numClusters(self.nframes)) * self.goal_init) ## Needs nframes to be set properly!!!!!!!!

        rewardtraj = np.arange(data_q.numTrajectories) # Recalculate reward for all states
        rewards = self.getRewards(rewardtraj, data_q, confstatdist, numstates, self.reward_method)
        for i in range(numstates):
            if len(rewards[i]) == 0:
                continue
            q_values[i] = updatingMean(q_values[i], n_values[i], rewards[i])
        n_values += np.array([len(x) for x in rewards])


        if self.save_qval:
            makedirs('saveddata', exist_ok=True)
            np.save(path.join('saveddata', 'e{}_qval.npy'.format(currepoch)), q_values)
            np.save(path.join('saveddata', 'e{}_nval.npy'.format(currepoch)), n_values)

        ucb_values = np.array([self.count_ucb(q_values[clust], self.exploration, currepoch + 1, n_values[clust]) for clust in range(numstates)])

        if self.save_qval:
            makedirs('saveddata', exist_ok=True)
            np.save(path.join('saveddata', 'e{}_ucbvals.npy'.format(currepoch)), ucb_values)

        N = self.nmax - self._running
        if self.actionpool <= 0:
            self.actionpool = N
       
        topactions = np.argsort(-ucb_values)[:self.actionpool]
        action = np.random.choice(topactions, N, replace=False)

        action_sel = np.zeros(numstates, dtype=int)
        action_sel[action] += 1
        while np.sum(action_sel) < N:  # When K is lower than N repeat some actions
            for a in action:
                action_sel[a] +=1
                if np.sum(action_sel) == N:
                    break

        if self.save_qval:
            np.save(path.join('saveddata', 'e{}_actions.npy'.format(currepoch)), action_sel)
        relFrames = self._getSpawnFrames_UCB(action_sel, data_q) 
        self._writeInputs(data.rel2sim(np.concatenate(relFrames)))
        return True

    def _getSimlist(self):
        from glob import glob
        from htmd.simlist import simlist, simfilter
        logger.info('Postprocessing new data')

        sims = simlist(glob(path.join(self.datapath, '*', '')), glob(path.join(self.inputpath, '*', '')),
                       glob(path.join(self.inputpath, '*', '')))

        if self.filter:
            sims = simfilter(sims, self.filteredpath, filtersel=self.filtersel)
        return sims


    def count_ucb(self, q_value, exploration, step, n_value):
        return (q_value + (exploration * np.sqrt((np.log(step) / (n_value + 1)))))

    def count_pucb(self, q_value, exploration, predictor, step, n_value):
        return (q_value + (exploration * predictor * np.sqrt((np.log(step) / (n_value + 1)))))

    def getRewards(self, trajidx, data_q, confstatdist, numstates, rewardmethod):
        from htmd.kinetics import Kinetics
        import pandas as pd
        rewards = [[] for _ in range(numstates)]
        for simidx in trajidx:
            # Get the eq distribution of each of the states the sim passed through
            states = data_q.St[simidx]
            statprob = confstatdist[simidx]
            connected = (states != -1) & (statprob != 0)
            if not np.any(connected):
                continue
            states = states[connected]
            statprob = statprob[connected]
            #energies = Kinetics._kB * self.temperature * np.log(statprob)
            energies = -Kinetics._kB * self.temperature * np.log(1-statprob)

            ww = len(energies)

            if rewardmethod == 'mean':
                windowedreward = pd.Series(energies[::-1]).rolling(ww, min_periods=1).mean().values[::-1]
            elif rewardmethod == 'max':
                windowedreward = pd.Series(energies[::-1]).rolling(ww, min_periods=1).max().values[::-1]
            else:
                raise RuntimeError('Reward method {} not available'.format(rewardmethod))

            for st, re in zip(states, windowedreward):
                rewards[st].append(re)

        return rewards

    def conformationStationaryDistribution(self, model):
        statdist = np.zeros(model.data.numFrames) # zero for disconnected set
        dataconcatSt = np.concatenate(model.data.St)
        for i in range(model.micronum):
            microframes = np.where(model.micro_ofcluster[dataconcatSt] == i)[0]
            statdist[microframes] = model.msm.stationary_distribution[i]
        return model.data.deconcatenate(statdist)

    def _checkNFrames(self, data):
        if self.nframes != 0 and data.numFrames >= self.nframes:
            logger.info('Reached maximum number of frames. Stopping adaptive.')
            return False
        return True

    def _getGoalData(self, sims):
        from htmd.projections.metric import Metric
        logger.debug('Starting projection of directed component')
        metr = Metric(sims, skip=self.skip)
        metr.set(self.goalfunction)
        data = metr.project()
        logger.debug('Finished calculating directed component')
        return data

    def _createMSM(self, data):
        from htmd.model import Model
        kmeanserror = True
        while kmeanserror:
            try:
                data.cluster(self.clustmethod(n_clusters=self._numClusters(data.numFrames)))
            except IndexError:
                continue
            kmeanserror = False
            
        self._model = Model(data)
        self._model.markovModel(self.lag, self._numMacrostates(data))
        if self.save:
            makedirs('saveddata', exist_ok=True)
            self._model.save(path.join('saveddata', 'e{}_adapt_model.dat'.format(self._getEpoch())))

    def _getSpawnFrames_UCB(self, reward, data):
        stateIdx = np.where(reward > 0)[0]
        _, relFrames = data.sampleClusters(stateIdx, reward[stateIdx], replacement=True, allframes=False)
        logger.debug('relFrames {}'.format(relFrames))
        return relFrames

    def _numClusters(self, numFrames):
        """ Heuristic that calculates number of clusters from number of frames """
        K = int(max(np.round(0.6 * np.log10(numFrames / 1000) * 1000 + 50), 100))  # heuristic
        if K > numFrames / 3:  # Ugly patch for low-data regimes ...
            K = int(numFrames / 3)
        return K

    def _numMacrostates(self, data):
        """ Heuristic for calculating the number of macrostates for the Markov model """
        macronum = self.macronum
        if data.K < macronum:
            macronum = np.ceil(data.K / 2)
            logger.warning('Using less macrostates than requested due to lack of microstates. macronum = ' + str(macronum))

        # Calculating how many timescales are above the lag time to limit number of macrostates
        from pyemma.msm import timescales_msm
        timesc = timescales_msm(data.St.tolist(), lags=self.lag, nits=macronum).get_timescales()
        macronum = min(self.macronum, max(np.sum(timesc > self.lag), 2))
        return macronum


import unittest
class _TestAdaptiveBandit(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from htmd.util import tempname
        from htmd.home import home
        from moleculekit.projections.metricdistance import MetricDistance
        import shutil
        import os

        tmpdir = tempname()
        shutil.copytree(home(dataDir='adaptive'), tmpdir)
        os.chdir(tmpdir)
        print(f"Running AdaptiveBandit test in {tmpdir}")

    def test_adaptive(self):
        from sklearn.cluster import MiniBatchKMeans
        from jobqueues.localqueue import LocalCPUQueue
        from moleculekit.projections.metricdistance import MetricDistance

        import numpy as np
        import random
        np.random.seed(0)  # Needed for the clustering to always give same results
        random.seed(0)

        md = AdaptiveBandit()
        md.app = LocalCPUQueue()
        md.generatorspath = 'generators'
        md.inputpath = 'input'
        md.datapath = 'data'
        md.coorname = 'input.coor'
        md.filter = True
        md.filtersel = 'all'

        md.clustmethod = MiniBatchKMeans
        md.projection = MetricDistance('protein resid 173 and name CA', 'resname BEN and name C1 C2 C3 C7')
        md.ticadim = 2
        md.nmin=1
        md.nmax=2
        md.nepochs= 9999
        md.nframes = 1000000

        md.reward_method = 'mean'
        md.exploration = 0.01
        md.actionspace = 'tica'
        md.actionpool = 0
        md.recluster = False

        md.save = True
        md.dryrun = True
        md.run()

if __name__ == '__main__':
    unittest.main(verbosity=2)


"""
HTMD provides a variety of protocols for setting up different types of simulations.
The most basic protocols include equilibration and production runs, both of which support flat-bottom potentials and
constraints for atoms.
"""
# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
__author__ = 'gianni'



from moleculekit.tools.preparation import *
import logging


logger = logging.getLogger(__name__)

logger.warning('Please do not import from htmd.builder.preparation. This will be deprecated. '
                'Use moleculekit.tools.preparation instead.')


# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import os
import sys
from subprocess import call
import inspect
import htmdx
import json
import urllib.request
import urllib.parse
import platform

has_connection = True


def htmd_do_register():
    do_register("htmd")


def show_news():
    try:
        response = urllib.request.urlopen(
            "https://www.htmd.org/news/content", timeout=3.05
        )
        print(response.read().decode("ascii"))
    except:
        pass


def check_approval(product, reg_file):
    # TODO: this may come back in other form
    # from htmdx.license import licenseEntitlements
    # jj = licenseEntitlements()
    # if "HTMD" in jj: return True
    # if "htmd" in jj: return True

    registration_data = {}
    try:
        with open(os.path.join(reg_file), "r") as f:
            registration_data = json.load(f)
    except:
        # Couldn't read registration file
        return False

    if not registration_data["ret"]:
        # There's no registration code in the file
        return False

    # Do online check
    try:
        payload = {"code": registration_data["ret"], "product": product}
        data = urllib.parse.urlencode(payload).encode("ascii")
        with urllib.request.urlopen(
            "https://www.acellera.com/licensing/htmd/check.php", data, timeout=3.05
        ) as f:
            ret = json.loads(f.read().decode("ascii"))

        if "approved" in ret:
            return True
        if "pending" in ret:
            return True
    except:
        return True

    return False


def check_registration(product=None):
    if not product:
        product = "NA"

    reg_file = os.path.join(
        os.path.expanduser("~"), ".htmd", ".registered-htmd", "registration"
    )

    if not (os.path.isfile(reg_file) and check_approval(product, reg_file)):
        accept_license(product=product)
        do_register(product=product)


def accept_license(product=None):
    show_license(product=product)
    print("Type 'yes' to accept the license [no] : ", end="")
    sys.stdout.flush()
    ret = input().strip()
    if ret != "yes":
        os._exit(1)


def show_license(product=None):
    path = os.path.abspath(
        os.path.dirname(os.path.dirname(inspect.getfile(accept_license)))
    )
    if product:
        path = os.path.join(path, product)

    path = os.path.join(path, "LICENCE.txt")
    if not os.path.exists(path):
        print("Could not find license file. Aborting")
        os._exit(1)

    print(sys.stdin)
    if sys.stdout.isatty() and sys.stdin:
        if platform.system() == "Windows":
            call(["c:\\Windows\\System32\\more.com", path])
        else:
            call(["more", path])
    else:
        fh = open(path, "r")
        print(fh.read())
        fh.close()


def do_register(product=None):
    name = ""
    institution = ""
    email = ""
    city = ""
    country = ""

    print(
        "\n Welcome to HTMD. We'd like to know who you are so that we can keep in touch!"
    )
    print(" Please enter your name, affiliation and institutional email address.\n")
    print(" Please provide valid information so that it might be verified.\n")

    while email == "" or not ("@" in email) or not ("." in email):
        print(" Institutional Email : ", end="")
        sys.stdout.flush()
        email = input().strip()

    while name == "":
        print(" Full name   : ", end="")
        sys.stdout.flush()
        name = input().strip()

    while institution == "":
        print(" Institution : ", end="")
        sys.stdout.flush()
        institution = input().strip()

    while city == "":
        print(" City        : ", end="")
        sys.stdout.flush()
        city = input().strip()

    while country == "":
        print(" Country     : ", end="")
        sys.stdout.flush()
        country = input().strip()

    payload = {
        "name": name,
        "institution": institution,
        "email": email,
        "product": product,
        "city": city,
        "country": country,
    }
    try:
        data = urllib.parse.urlencode(payload).encode("ascii")
        with urllib.request.urlopen(
            "https://www.acellera.com/licensing/htmd/register.php", data
        ) as f:
            text = f.read().decode("ascii")

        prefix = os.path.join(
            os.path.expanduser("~"), ".htmd", ".registered-" + product
        )
        os.makedirs(prefix, exist_ok=True)
        regfile = os.path.join(prefix, "registration")
        with open(regfile, "w") as fh:
            fh.write(text)
        print("")
    except:
        print("\nCould not register!\n")


def main_activate():
    import sys

    if len(sys.argv) != 2:
        print(
            "\n Acellera License Installation\n\n Syntax: activate [activation token]\n\nPlease contact support@acellera.com for licensing\n\n"
        )
        sys.exit(0)

    token = sys.argv[1]
    response = urllib.request.urlopen(
        "https://licensing.acellera.com/issue/?token=" + token
    )
    text = response.read()
    if isinstance(text, bytes):
        text = text.decode("ascii")

    if response.status != 200:
        print("\n Error: " + text + "(" + str(response.status) + ")\n\n")
    else:
        prefix = os.path.join(os.path.expanduser("~"), ".acellera")
        try:
            os.mkdir(prefix)
        except:
            pass

        with open(os.path.join(prefix, "license.dat"), "a") as f:
            f.write(text)

        print("\n License installed in ~/.acellera/license.dat\n\n")
        try:
            with open(os.path.join("/opt/acellera/license.dat"), "a") as f:
                f.write(text)
            print("\n License installed in /opt/acellera/license.dat\n\n")
        except:
            pass


if __name__ == "__main__":
    main_activate()



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import ctypes as ct
import os
import platform
import htmdx
from htmd.decorators import _Deprecated


@_Deprecated('1.13.5')
def licenseEntitlements():
    libdir = os.path.join(htmdx.__path__[0], "..", "htmd", "lib", platform.system())

    # No liblicense.so for OS X yet. Need to buy it
    if platform.system() == "Darwin":
        return dict()

    if platform.system() == "Windows":
        ct.cdll.LoadLibrary(os.path.join(libdir, "libgcc_s_seh-1.dll"))
        if os.path.exists(os.path.join(libdir, "psprolib.dll")):
            ct.cdll.LoadLibrary(os.path.join(libdir, "psprolib.dll"))
        lib = ct.cdll.LoadLibrary(os.path.join(libdir, "liblicense.dll"))
    else:
        lib = ct.cdll.LoadLibrary(os.path.join(libdir, "liblicense.so"))
    arg1 = ct.create_string_buffer(2048)
    arg2 = ct.c_int(2048)
    lib.license_check(arg1, arg2)
    val = (arg1.value.decode("utf-8")).split(":")

    toks = dict()
    for v in val:
        x = v.split("-")
        if len(x) > 1:
            if x[0] not in toks:
                toks[x[0]] = dict()
            if x[1] not in toks[x[0]]:
                toks[x[0]][x[1]] = 0
            try:
                toks[x[0]][x[1]] += int(x[2])
            except:
                pass
    return toks


if __name__ == "__main__":
    print("License entitlements:")
    print(licenseEntitlements())



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import numpy as np
import logging
logger = logging.getLogger(__name__)


class ModelHMM(object):
    def __init__(self, data=None):
        if data is None:
            return
        self.data = data

    def markovModel(self, lag, macronum):
        from pyemma.msm import estimate_hidden_markov_model
        self.hmm = estimate_hidden_markov_model(self.data.St.tolist(), macronum, lag, connectivity='largest')
        print('Active set includes macrostates: {}'.format(self.hmm.active_set))

    @property
    def macronum(self):
        return len(self.hmm.active_set)

    def plotTimescales(self, maxlag, numstates):
        import pyemma.msm as msm
        import pyemma.plots as mplt
        its = msm.timescales_hmsm(self.data.St.tolist(), numstates, lags=maxlag)
        mplt.plot_implied_timescales(its, ylog=True, units='ns', dt=self.data.fstep, linewidth=2)

    def eqDistribution(self, plot=True):
        if plot:
            from matplotlib import pylab as plt
            plt.bar(range(self.macronum), self.hmm.stationary_distribution)
            #ax = plt.gca()
            #ax.set_xticklabels([str(a) for a in self.hmm.active_set])
            plt.xticks(np.arange(self.macronum)+0.4, [str(a) for a in self.hmm.active_set])
        return self.hmm.stationary_distribution

    def viewStates(self, protein=None, ligand=None, nsamples=20):
        from htmd.projections.metric import _singleMolfile
        from moleculekit.molecule import Molecule
        from moleculekit.vmdviewer import getCurrentViewer
        (single, molfile) = _singleMolfile(self.data.simlist)
        if not single:
            raise RuntimeError('Can''t visualize states without unique molfile')

        viewer = getCurrentViewer()
        colors = [0, 1, 3, 4, 5, 6, 7, 9]

        print('Active set includes macrostates: {}'.format(self.hmm.active_set))

        # dtraj = np.vstack(self.hmm.discrete_trajectories_full)
        res = self.hmm.sample_by_observation_probabilities(nsamples)
        refmol = Molecule(molfile)

        for i, s in enumerate(self.hmm.active_set):
            mol = Molecule(molfile)
            mol.coords = []
            mol.box = []
            # idx = np.where(dtraj == i)[0]
            # samples = np.random.choice(idx, 20)
            # frames = self.data.abs2sim(samples)

            frames = self.data.rel2sim(res[i])
            for f in frames:
                mol._readTraj(f.sim.trajectory[f.piece], frames=[f.frame], append=True)
            mol.wrap('protein')
            mol.align('protein', refmol=refmol)
            viewer.loadMol(mol, name='hmm macro ' + str(s))
            if ligand is not None:
                viewer.rep('ligand', sel=ligand, color=colors[np.mod(i, len(colors))])
            if protein is not None:
                viewer.rep('protein')
            viewer.send('start_sscache')



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import os
from moleculekit.molecule import Molecule
import numpy as np
import sys
import logging
logger = logging.getLogger(__name__)


def solvate(mol, pad=None, minmax=None, negx=0, posx=0, negy=0, posy=0, negz=0, posz=0, buffer=2.4, watsize=65.4195,
            prefix='WT', keysel='name OH2', rotate=False, rotsel='all', rotinc=36, spdb=None,
            spsf=None, stop=None):
    """ Solvates the system in a water box


    Parameters
    ----------
    mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The molecule object we want to solvate
    pad : float
        The padding to add to the minmax in all dimensions. You can specify different padding in each dimension using
        the negx, negy, negz, posx, posy, posz options. This option will override any values in the neg and pos options.
    minmax : list
        Min and max dimensions. Should be a 2D matrix of the form [[minx, miny, minz], [maxx, maxy, maxz]]. If none is
        given, it is calculated from the minimum and maximum coordinates in the mol.
    negx : float
        The padding in the -x dimension
    posx : float
        The padding in the +x dimension
    negy : float
        The padding in the -y dimension
    posy : float
        The padding in the +y dimension
    negz : float
        The padding in the -z dimension
    posz : float
        The padding in the +z dimension
    buffer : float
        How much buffer space to leave empty between waters and other molecules
    watsize : float
        The size of the water box
    prefix : str
        The prefix used for water segments
    keysel : str
        The key selection for water atoms
    rotate : bool
        Enable automated rotation of molecule to fit best in box
    rotsel : str
        The selection of atoms to rotate
    rotinc : float
        The increment in degrees to rotate
    spdb : str
        The path to the water pdb file
    spsf : str
        The path to the water psf file
    stop : str
        The path to the water topology file

    Returns
    -------
    mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        A solvated molecule

    Examples
    --------
    >>> smol = solvate(mol, pad=10)
    >>> smol = solvate(mol, minmax=[[-20, -20, -20],[20, 20, 20]])
    """
    from tqdm import tqdm
    from htmd.home import home

    mol = mol.copy()
    if mol.numFrames > 1:
        logger.warning('Multiple frames in Molecule. Solvate keeps only frame 0 and discards the rest.')
        mol.coords = np.atleast_3d(mol.coords[:, :, 0])

    if spdb is None:
        spdb = os.path.join(home(shareDir=True), 'solvate', 'wat.pdb')

    if os.path.isfile(spdb):
        logger.info('Using water pdb file at: ' + spdb)
        water = Molecule(spdb)
    else:
        raise NameError('No solvent pdb file found in ' + spdb)

    if pad is not None:
        negx = pad; posx = pad; negy = pad; posy = pad; negz = pad; posz = pad

    if rotate:
        raise NameError('Rotation not implemented yet')

    # Calculate min max coordinates from molecule
    if mol.numAtoms > 0:
        minmol = np.min(mol.get('coords'), axis=0)
        maxmol = np.max(mol.get('coords'), axis=0)
    else:
        minmol = [np.inf, np.inf, np.inf]
        maxmol = [-np.inf, -np.inf, -np.inf]

    if minmax is None:
        minc = minmol
        maxc = maxmol
    else:
        if isinstance(minmax, list):
            minmax = np.array(minmax)
        minc = minmax[0, :]
        maxc = minmax[1, :]

    xmin = float(minc[0] - negx)
    xmax = float(maxc[0] + posx)
    ymin = float(minc[1] - negy)
    ymax = float(maxc[1] + posy)
    zmin = float(minc[2] - negz)
    zmax = float(maxc[2] + posz)

    dx = xmax - xmin
    dy = ymax - ymin
    dz = zmax - zmin

    nx = int(np.ceil((dx + 2 * buffer) / watsize))
    ny = int(np.ceil((dy + 2 * buffer) / watsize))
    nz = int(np.ceil((dz + 2 * buffer) / watsize))

    # Calculate number of preexisting water segments with given prefix
    if mol.numAtoms > 0:
        preexist = len(np.unique(mol.get('segid', sel='segid "{}.*"'.format(prefix))))
    else:
        preexist = 0

    numsegs = nx * ny * nz
    logger.info('Replicating ' + str(numsegs) + ' water segments, ' + str(nx) + ' by ' + str(ny) + ' by ' + str(nz))

    # Check that we won't run out of segment name characters, and switch to
    # using hexadecimal or alphanumeric naming schemes in cases where decimal
    # numbered segnames won't fit into the field width.
    testsegname    = '{0}{1:d}'.format(prefix, numsegs + preexist)
    testsegnamehex = '{0}{1:X}'.format(prefix, numsegs + preexist)
    writemode = 'decimal'
    if len(testsegname) > 4 and len(testsegnamehex) <= 4:
        writemode = 'hex'
        logger.warning('Warning: decimal naming would overrun segname field. Using hexadecimal segnames instead...')
    elif len(testsegnamehex) > 4:
        writemode = 'alphanum'
        logger.warning('Warning: decimal or hex naming would overrun segname field. Using alphanumeric segnames instead...')

    minx = minmol[0] - buffer; miny = minmol[1] - buffer; minz = minmol[2] - buffer
    maxx = maxmol[0] + buffer; maxy = maxmol[1] + buffer; maxz = maxmol[2] + buffer

    bar = tqdm(total=nx*ny*nz, desc='Solvating')
    waterboxes = np.empty(numsegs, dtype=object)
    n = preexist
    w = 0
    for i in range(nx):
        movex = xmin + i * watsize
        movexmax = movex + watsize
        xoverlap = True
        if movex > maxx or movexmax < minx:
            xoverlap = False

        for j in range(ny):
            movey = ymin + j * watsize
            moveymax = movey + watsize
            yoverlap = True
            if movey > maxy or moveymax < miny:
                yoverlap = False

            for k in range(nz):
                movez = zmin + k * watsize
                movezmax = movez + watsize
                zoverlap = True
                if movez > maxz or movezmax < minz:
                    zoverlap = False

                if writemode == 'decimal':
                    segname = '{0}{1:d}'.format(prefix, n)
                elif writemode == 'hex':
                    segname = '{0}{1:x}'.format(prefix, n)
                elif writemode == 'alphanum':
                    segname = '{0}{1:c}{2:c}{3:c}'.format(prefix, int(np.floor(np.floor(n/26)/26) + 65), int(np.mod(np.floor(n/26), 26) + 65), int(np.mod(n, 26) + 65))

                waterboxes[w] = water.copy()
                waterboxes[w].moveBy([movex, movey, movez])
                waterboxes[w].set('segid', segname)

                mol.append(waterboxes[w])
                watsel = mol.segid == segname

                selover = np.zeros(len(watsel), dtype=bool)
                if xoverlap and yoverlap and zoverlap:  # Remove water overlapping with other segids
                    selover = _overlapWithOther(mol, segname, buffer)
                # Remove water outside the boundaries
                selout = _outOfBoundaries(mol, segname, xmin, xmax, ymin, ymax, zmin, zmax)
                sel = selover | selout

                #mol.write('temp.pdb')
                mol.filter(mol.segid != segname, _logger=False)
                waterboxes[w].filter(np.invert(sel[watsel]), _logger=False)
                #waterboxes[w].write('wat' + str(w) + '.pdb')
                n += 1
                w += 1
                bar.update(1)
    bar.close()

    waters = 0
    for i in range(numsegs):
        waters += waterboxes[i].numAtoms
        if waterboxes[i].numAtoms != 0:
            mol.append(waterboxes[i])

    logger.info('{} water molecules were added to the system.'.format(int(waters/3)))
    return mol


def _overlapWithOther(mol, segname, buffer):
    sel = 'segid {} and same resid as (segid {} and within {} of not segid {})'.format(segname, segname, buffer, segname)
    return mol.atomselect(sel)


def _outOfBoundaries(mol, segname, xmin, xmax, ymin, ymax, zmin, zmax):
    sel = 'segid {} and same resid as (segid {} and (x < {} or x > {} or y < {} or y > {} or z < {} or z > {}))'.format(
        segname, segname, xmin, xmax, ymin, ymax, zmin, zmax
    )
    return mol.atomselect(sel)



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from htmd.version import version as _version
from htmd.versionwarnings import _issueWarnings, _disableWarnings
import __main__ as main
import os.path
from htmd.config import _config
import htmd.home
import logging.config

__version__ = _version()

try:
    logging.config.fileConfig(os.path.join(htmd.home.home(), 'logging.ini'), disable_existing_loggers=False)
except:
    print("HTMD: Logging setup failed")

reg_file = os.path.join(os.path.expanduser('~'), '.htmd', '.registered-htmd', 'registration')
if (not hasattr(main, '__file__')) and \
        ((not os.path.isfile(reg_file)) or os.getenv("LICENCE_ACCEPTED") == "YES" or os.getenv("TRAVIS_REPO_SLUG")):
    print('\nCopyright by Acellera Ltd. By executing you are accepting the License. In order to register, '
          'run htmd_register on your terminal.')
    print('The registration information must be valid so that it might be verified.')

if _config['configfile']:
    if not os.path.isfile(_config['configfile']):
        raise FileNotFoundError('HTMD Config file {} does not exist.'.format(_config['configfile']))
    elif not _config['configfile'].endswith('.py'):
        raise Warning('HTMD Config file {} may not be a python file.'.format(_config['configfile']))
    else:
        try:
            with open(_config['configfile']) as f:
                exec(compile(f.read(), _config['configfile'], 'exec'))
        except:
            raise RuntimeError('Failed to execute the HTMD Config file {}.'.format(_config['configfile']))
        else:
            print('\nHTMD Config file {} executed.'.format(_config['configfile']))

_issueWarnings()


from moleculekit.molecule import Molecule
import logging

logger = logging.getLogger(__name__)

logger.warning('Please do not import htmd.molecule.molecule. It has been moved to the moleculekit package. This import will be deprecated.')


# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import simtk.openmm as mm
from sys import stdout
from simtk import unit
from moleculekit.molecule import Molecule
from simtk.openmm import app


defaultCharmmFiles = [
    'top_all36_prot.rtf',
    'par_all36_prot.prm',
    'top_all36_na.rtf',
    'par_all36_na.prm',
    'top_all36_carb.rtf',
    'par_all36_carb.prm',
    'top_all36_lipid.rtf',
    'par_all36_lipid.prm',
    'top_all36_cgenff.rtf',
    'par_all36_cgenff.prm',
    'toppar_all36_prot_retinol.str',
    'toppar_all36_na_rna_modified.str',
    'toppar_all36_carb_glycopeptide.str',
    'toppar_all36_prot_fluoro_alkanes.str',
    'toppar_all36_prot_na_combined.str',
    'toppar_all36_prot_heme.str',
    'toppar_all36_lipid_bacterial.str',
    'toppar_all36_lipid_miscellaneous.str',
    'toppar_all36_lipid_cholesterol.str',
    'toppar_all36_lipid_yeast.str',
    'toppar_all36_lipid_sphingo.str',
    'toppar_all36_lipid_inositol.str',
    'toppar_all36_lipid_cardiolipin.str',
    'toppar_all36_lipid_detergent.str',
    'toppar_all36_lipid_lps.str',
    'toppar_water_ions.str',
    'toppar_dum_noble_gases.str',
    'toppar_all36_na_nad_ppi.str',
    'toppar_all36_carb_glycolipid.str',
    'toppar_all36_carb_imlab.str'
]


def _readCharmmParameters(folder, listfile=None):
    from glob import glob
    import os
    if listfile is None:
        files = glob(os.path.join(folder, '*'))
    elif isinstance(listfile, list):
        files = [os.path.join(folder, f) for f in listfile]
    return app.CharmmParameterSet(*files)


def _printEnergies(sim, text):
    state = sim.context.getState(getEnergy=True)
    pot = state.getPotentialEnergy()
    kin = state.getKineticEnergy()
    print("{} {} {} {}".format(text, pot + kin, pot, kin))


def _getPlatform(plat, device):
    platform = mm.Platform.getPlatformByName(plat)
    prop = None
    if plat == 'CUDA':
        prop = {'CudaPrecision': 'single', 'CudaDeviceIndex': str(device)}
    return platform, prop


def equilibrateSystem(pdbfile, psffile, outpdb, numsteps=30000, minimplatform='CPU', equilplatform='CUDA', device=0,
                      temp=300, minimize=500000, minimizetol=100, charmmfolder=None):
    pdb = Molecule(pdbfile)
    watcoo = pdb.get('coords', 'water')
    celld = unit.Quantity((watcoo.max(axis=0) - watcoo.min(axis=0)).squeeze(), unit=unit.angstrom)

    psf = app.CharmmPsfFile(psffile)
    pdb = app.PDBFile(pdbfile)
    psf.setBox(celld[0], celld[1], celld[2])

    params = _readCharmmParameters(charmmfolder, defaultCharmmFiles)

    system = psf.createSystem(params, nonbondedMethod=app.PME,
                              nonbondedCutoff=1*unit.nanometer,
                              constraints=app.HBonds)
    system.addForce(mm.MonteCarloBarostat(1*unit.atmospheres, temp*unit.kelvin, 25))

    platform, prop = _getPlatform(minimplatform, device)
    integrator = mm.LangevinIntegrator(temp*unit.kelvin, 1/unit.picosecond, 0.002*unit.picoseconds)
    simulation = app.Simulation(psf.topology, system, integrator, platform, prop)
    simulation.context.setPositions(pdb.positions)

    # Perform minimization
    _printEnergies(simulation, 'Energy before minimization')
    simulation.minimizeEnergy(tolerance=minimizetol*unit.kilojoule/unit.mole, maxIterations=minimize)
    _printEnergies(simulation, 'Energy after minimization')

    # Copy coordinates from miminization context to equilibration context
    state = simulation.context.getState(getPositions=True)
    pos = state.getPositions()

    # Set up the equilibration simulation
    platform, prop = _getPlatform(equilplatform, device)
    integrator = mm.LangevinIntegrator(temp*unit.kelvin, 1/unit.picosecond, 0.002*unit.picoseconds)
    simulation = app.Simulation(psf.topology, system, integrator, platform, prop)
    simulation.context.setPositions(pos)

    # Generate initial velocities
    simulation.context.setVelocitiesToTemperature(temp)

    from htmd.membranebuilder.pdbreporter import PDBReporter
    simulation.reporters.append(PDBReporter(outpdb, numsteps, enforcePeriodicBox=False))
    simulation.reporters.append(app.StateDataReporter(stdout, int(numsteps/10), step=True,
        potentialEnergy=True, kineticEnergy=True, totalEnergy=True, temperature=True,
        progress=True, speed=True, remainingTime=True, totalSteps=numsteps, separator='\t'))
    simulation.step(numsteps)




# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#

#import dill
import pickle
import logging
logger = logging.getLogger(__name__)


def htmdsave(fname, varnames=None):
    """ Saves the workspace to a file

    Parameters
    ----------
    fname : str
        The file in which to save the workspace
    varnames : list
        A list of variables to save. Default: saves all variables.
    """
    from types import ModuleType, FunctionType
    import __main__ as _main_module
    import inspect
    mydict = dict()
    maindict = _main_module.__dict__.copy()
    bannedvars = ['In', 'Out', 'exit', 'quit', 'get_ipython']

    if varnames is None:
        logger.warning('Save will ignore all variables starting with _ and following variable names {}'.format(bannedvars))
    logger.warning('Save will duplicate data which is stored in various pointers or variables. Take care!')

    if varnames is None:
        for k in maindict:
            if k[0] != '_' and k not in bannedvars and \
                    not isinstance(maindict[k], ModuleType) and \
                    not isinstance(maindict[k], FunctionType) and \
                    not inspect.isclass(maindict[k]):
                try:
                    pickle.dumps(maindict[k])  # This probably slows down everything
                    mydict[k] = maindict[k]
                except:
                    logger.debug('Skipping variable: {}'.format(k))
    else:
        if not isinstance(varnames, list):
            varnames = [varnames]
        for k in varnames:
            mydict[k] = maindict[k]

    logger.debug('Saving variables: {}'.format(list(mydict.keys())))
    f = open(fname, 'wb')
    pickle.dump(mydict, f)
    f.close()


def htmdload(fname):
    """ Loads a previously saved workspace from a file.

    Parameters
    ----------
    fname : str
        The location of the file containing the saved workspace.
    """
    import __main__ as _main_module

    f = open(fname, 'rb')
    mydict = pickle.load(f)
    f.close()

    # Inserting all variables into the main dictionary
    _main_module.__dict__.update(mydict)


'''def save(fname):
    from htmd.metricdata import MetricData
    from htmd.model import Model
    from htmd.kinetics import Kinetics

    import __main__ as _main_module
    myobj = dict()
    for k in _main_module.__dict__:
        if isinstance(_main_module.__dict__[k], MetricData) or isinstance(_main_module.__dict__[k], Model):
            myobj[k] = _main_module.__dict__[k]
        # TODO: Support lists and np.arrays of class objects

    f = open(fname, 'wb')
    pickle.dump(myobj, f)
    f.close()
    #dill.dump_session(filename=filename)
'''

'''
def save(dir)
    from htmd.metricdata import MetricData
    from htmd.model import Model
    from htmd.kinetics import Kinetics

    if not path.isdir(dir):
        os.makedirs(dir)

    import __main__ as _main_module
    myobj = dict()
    for k in _main_module.__dict__:
        if isinstance(_main_module.__dict__[k], MetricData):
            _main_module.__dict__[k].save(path.join(dir, 'MetricData_{}.dat'.format(k)))
        elif isinstance(_main_module.__dict__[k], Model):
            _main_module.__dict__[k].save(path.join(dir, 'Model_{}.dat'.format(k)))

def load(dir):
    from htmd.metricdata import MetricData
    from htmd.model import Model
    from htmd.kinetics import Kinetics
    from glob import glob

    if not path.isdir(dir):
        raise NameError('Directory {} does not exist'.format(dir))

    import __main__ as _main_module

    regex = re.compile('(\w+?)_(\w+).dat')

    datfiles = glob(path.join(dir, '*.dat'))
    for d in datfiles:
        res = regex.search(path.basename(d))
        if not res:  # If we are running on top of adaptive, use the first name part for the next sim name
            continue

        classname = res.group(1)
        varname = res.group(2)
        print(classname, varname)

        if classname == 'MetricData':
            _main_module.__dict__[varname] = MetricData()
        elif classname == 'Model':
            _main_module.__dict__[varname] = Model()
        _main_module.__dict__[varname].load(d)'''



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import numpy as np
from numba import jit
from math import sqrt, atan2


@jit(nopython=True)
def cross(vec1, vec2):
    """ Calculate the dot product of two 3d vectors. """
    a1, a2, a3 = vec1[0], vec1[1], vec1[2]
    b1, b2, b3 = vec2[0], vec2[1], vec2[2]
    result = np.zeros(3)
    result[0] = a2 * b3 - a3 * b2
    result[1] = a3 * b1 - a1 * b3
    result[2] = a1 * b2 - a2 * b1
    return result


@jit(nopython=True)
def dot(vec1, vec2):
    return vec1[0] * vec2[0] + vec1[1] * vec2[1] + vec1[2] * vec2[2]


@jit(nopython=True)
def norm(vec):
    n = 0
    for i in range(3):
        n += vec[i] * vec[i]
    return sqrt(n)


@jit(nopython=True)
def pairwiseRMSD(coords):
    nframes = coords.shape[2]
    pairwise = np.zeros(int(nframes*(nframes-1)/2), dtype=np.float32)

    k = 0
    for i in range(nframes):
        for j in range(i+1, nframes):
            pairwise[k] = np.sqrt(np.mean((coords[:, :, i] - coords[:, :, j]) ** 2))
            k +=1
    return pairwise



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import os
import numpy as np
from protocolinterface import ProtocolInterface, val
from htmd.util import ensurelist
import logging
from unittest import TestCase

logger = logging.getLogger(__name__)


class _Restraint:
    def __init__(self, contype, selection, width, restraints, axes='xyz', fbcentre=None, fbcentresel=None):
        self.type = contype
        self.selection = selection
        self.width = ensurelist(width)
        self.restraints = restraints
        self.axes = axes
        self.fbcentre = fbcentre
        self.fbcentresel = fbcentresel

        if len(self.width) != 1 and len(self.width) != 3:
            raise RuntimeError('Restraint width must be either a single value or a list of 3 values for the xyz '
                               'dimensions')
        if self.fbcentre is not None:
            self.fbcentre = ensurelist(self.fbcentre)
            if len(self.fbcentre) != 3:
                raise RuntimeError('Restraint fbcentre must be a list of 3 values for the xyz coordinates')

    def format(self, maxwidth=None):
        if maxwidth is None:
            res = '{type}Restraint '.format(type=self.type)
        else:
            res = '{type: <{maxwidth}}\t'.format(type=self.type+'Restraint', maxwidth=maxwidth)

        res += '"{}" '.format(self.selection)

        if self.fbcentre is not None:
            res += 'fbcentre "{}" '.format(' '.join(map(str, self.fbcentre)))
        if self.fbcentresel is not None:
            res += 'fbcentresel "{}" '.format(self.fbcentresel)

        res += 'axes {axes} width "{width}" '.format(axes=self.axes, width=' '.join(map(str, self.width)))
        res += 'setpoints'
        for p in self.restraints:
            res += ' {rest}@{time}'.format(rest=p[0], time=p[1])
        return res

    def __str__(self):
        return self.format()

    @staticmethod
    def _fromDict(redict):
        if 'axes' not in redict:
            redict['axes'] = None
        if 'fbcentre' not in redict:
            redict['fbcentre'] = None
        if 'fbcentresel' not in redict:
            redict['fbcentresel'] = None

        return _Restraint(redict['type'], redict['selection'], redict['width'], redict['restraints'], redict['axes'],
                          redict['fbcentre'], redict['fbcentresel'])

    def _toDict(self):
        return self.__dict__


class GroupRestraint(_Restraint):
    def __init__(self, selection, width, restraints, fbcentre=None, fbcentresel=None, axes='xyz'):
        """ A restraint applied on the center of mass of a group of atoms.

        Parameters
        ----------
        selection : str
            Atom selection string of the atoms on whose center of mass the restraints will be applied.
            See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
        width : float or list of floats
            The width of the flat-bottom potential box in Angstrom
        restraints : list of tuples
            A list of restraint/time pairs indicating which value the restraints should have at each point in time.
            Restraints are given in kcal/mol units. Time can be either a timestep or time if one of the suffices
            "us", "ns", "ps", "fs" is used.
        axes : str
            The axes on which to apply the potential.
        fbcentre : list of floats
            You can set an explicit center for the FB potential using a list of xyz coordinates.
        fbcentresel : str
            You can select atoms whose center of mass will be used as the center of the FB potential.

        Examples
        --------
        >>> gr = GroupRestraint('resname MOL', 10, [(20, '10ns'), (10, '20ns'), (0, '30ns')])
        """
        super().__init__('group', selection, width, restraints, axes, fbcentre=fbcentre, fbcentresel=fbcentresel)


class AtomRestraint(_Restraint):
    def __init__(self, selection, width, restraints, axes='xyz'):
        """ A restraint applied individually on each atom in the selection

        Parameters
        ----------
        selection : str
            Atom selection string of the atoms on whose center of mass the restraints will be applied.
            See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
        width : float or list of floats
            The width of the flat-bottom potential box in Angstrom
        restraints : list of tuples
            A list of restraint/time pairs indicating which value the restraints should have at each point in time.
            Restraints are given in kcal/mol units. Time can be either a timestep or time if one of the suffices
            "us", "ns", "ps", "fs" is used.
        axes : str
            The axes on which to apply the potential.

        Examples
        --------
        >>> gr = AtomRestraint('name CA', 10, [(20, '10ns'), (10, '20ns'), (0, '30ns')])
        """
        super().__init__('atom', selection, width, restraints, axes)


class _Acemd(ProtocolInterface):
    _defaultfnames = {'bincoordinates': 'input.coor', 'binvelocities': 'input.vel', 'binindex': 'input.idx',
                      'structure': 'structure.*', 'parameters': 'parameters', 'extendedsystem': 'input.xsc',
                      'coordinates': 'structure.pdb', 'velocities': 'velocity.pdb', 'consref': 'structure.pdb',
                      'parmfile': 'parameters'}

    def __init__(self, version=3):
        super().__init__()
        self._version = version
        self._file_data = {}
        self._outnames = {}

    def _amberConfig(self):
        # AMBER specific fixes
        if self.structure.endswith('.prmtop'):
            if self.parmfile is None:
                self.parmfile = self.parameters
            self.parameters = None

    def load(self, path='.'):
        """ Loads all files required to run a simulation and apply eventually configured protocols to it

        Parameters
        ----------
        path : str
            Working directory relative to which the configuration file is read
        """
        # load files and reset filenames

        for cmd in self._defaultfnames.keys():
            if cmd in self.__dict__ and self.__dict__[cmd] is not None:
                found = False

                for fname in ensurelist(self.__dict__[cmd]):
                    fpath = os.path.join(path, fname)
                    if not os.path.exists(fpath):
                        continue

                    f = open(fpath, 'rb')  # read all as binary
                    data = f.read()
                    f.close()
                    self._file_data[cmd] = data

                    defaultname = self._defaultfnames[cmd]
                    if defaultname.endswith('*'):
                        defaultname = '{}.{}'.format(os.path.splitext(defaultname)[0], os.path.splitext(fpath)[1][1:])
                    self._outnames[cmd] = defaultname
                    self.__dict__[cmd] = fname
                    found = True
                    break
                if not found:
                    raise RuntimeError('Could not find any of the files "{}" specified for command "{}" '
                                       'in path {}'.format(self.__dict__[cmd], cmd, path))

        self._amberConfig()  # Change stuff for AMBER
        if self._version == 3 and self.thermostattemperature is None:
            self.thermostattemperature = self.temperature

    def save(self, path, overwrite=False):
        """ Create a directory with all necessary input to run acemd.

        Parameters
        ----------
        path : str
            New execution directory to create
        overwrite : bool
            Overwrite output directory if it exists.
        """
        if self.run is None:
            raise RuntimeError('You need to specify the simulation length with the "run" command.')
        if self.temperature is None:
            raise RuntimeError('You need to specify the simulation temperature with the "temperature" command.')

        if os.path.exists(path):
            if overwrite:
                import shutil
                shutil.rmtree(path)
            else:
                raise RuntimeError('Directory exists, use overwrite=True or remove directory')
        os.makedirs(path)

        # Write all files using default filenames
        for f in self._file_data:
            fo = open(os.path.join(path, self._outnames[f]), 'wb')  # write all as binary
            fo.write(self._file_data[f])
            fo.close()

        self._writeBashRun(os.path.join(path, 'run.sh'))

        self._file_data = {}  # empty file dictionary after writing

        self.writeConf(os.path.join(path, 'input'))

    def setup(self, indir='.', outdir='run', overwrite=False):
        """ Convenience method performing load and save.

        Parameters
        ----------
        indir : str
            The directory to load from.
        outdir : str
            The directory to save to.
        overwrite : bool
            Overwrite output directory if it exists.
        """
        self.load(indir)
        self.save(outdir, overwrite)

    def show(self, quiet=False):
        """ Returns the Acemd configuration file string

        Parameters
        ----------
        quiet : bool
            If true it prints the string to stdout

        Returns
        -------
        conf : str
            The string of the configuration file
        """
        text = ''
        if 'TCL' in self.__dict__ and self.__dict__['TCL'] is not None:
            text = self.__dict__['TCL']

        text += '#\n'

        maxwidth = np.max([len(k) for k in self.__dict__.keys()])

        keys = sorted(list(self.__dict__.keys()))
        if 'restraints' in keys:
            keys += [keys.pop(keys.index('restraints'))]
        keys += [keys.pop(keys.index('run'))]  # Move the run command to the end
        for cmd in keys:
            if cmd == 'restraints' and self.restraints is not None:
                for r in ensurelist(self.restraints):
                    text += '{}\n'.format(r.format(maxwidth))
            elif not cmd.startswith('_') and self.__dict__[cmd] is not None and cmd != 'TCL':
                val = self.__dict__[cmd]
                if cmd in self._outnames:
                    val = self._outnames[cmd]
                name = cmd
                if cmd == 'scaling14':  # variables cannot start with numbers. We need to rename it here for acemd
                    name = '1-4scaling'
                text += '{name: <{maxwidth}}\t{val:}\n'.format(name=name, val=val, maxwidth=maxwidth)

        if not quiet:
            print(text)
        else:
            return text

    def _writeBashRun(self, fname):
        with open(fname, 'w') as f:
            if self._version == 3:
                f.write('#!/bin/bash\nacemd3 >log.txt 2>&1')
            elif self._version == 2:
                f.write('#!/bin/bash\nacemd >log.txt 2>&1')
            else:
                raise ValueError('Acemd version {} is not valid'.format_map(self._version))
        os.chmod(fname, 0o700)

    def __repr__(self):
        return self.show(quiet=True)

    def writeConf(self, fname='input'):
        """ Write an acemd configuration file

        Parameters
        ----------
        fname : output file name
        """
        text = self.show(quiet=True)
        fo = open(fname, 'w')
        fo.write(text)
        fo.close()


class Acemd2(_Acemd):
    """ Class for legacy support of ACEMD older versions ("version 2").

    Parameters
    ----------

    temperature : float, default=None
        Temperature of the thermostat in Kelvin.
    restart : ('on', 'off'), str, default=None
        Restart simulation.
    restartfreq : int, default=None
        Restart file frequency.
    outputname : str, default=None
        Output file name.
    xtcfile : str, default=None
        Output XTC file name.
    xtcfreq : int, default=None
        XTC sampling frequency in steps.
    timestep : float, default=None
        Simulation timestep.
    rigidbonds : ('none', 'all'), str, default=None
        Enable holonomic constraints on all hydrogen-heavy atom bond terms
    hydrogenscale : float, default=None
        Amount by which to scale the mass of H atoms
    switching : ('on', 'off'), str, default=None
        Enable to apply smoothing and switching functions to electrostatic and VdW forces.
    switchdist : float, default=None
        Range beyond which to begin to apply the switching functions.
    cutoff : str, default=None
        Cutoff distance for direct-space electrostatic interaction evaluation.
    exclude : ('none', '1-2', '1-3', '1-4', 'scaled1-4'), str, default=None
        Which pairs of bonded atoms should be excluded from non-bonded interactions
    scaling14 : float, default=None
        Scaling factor for 1-4 electrostatic interations.
    langevin : ('on', 'off'), str, default=None
        Enable the Langevin thermostat.
    langevintemp : float, default=None
        The set point in K for the Langevin thermostat.
    langevindamping : float, default=None
        Langevin damping constant gamma (1/ps)
    pme : ('on', 'off'), str, default=None
        Enable the use of PME for long-range electrostatics.
    pmegridspacing : float, default=None
        The spacing of the PME mesh in 1/A.
    fullelectfrequency : int, default=None
        The frequency in interations between successive calculations of long-range (PME) electrostatics.
    energyfreq : int, default=None
        The frequency with which ACEMD will calculate system energies.
    constraints : ('on', 'off'), str, default=None
        Set to enable positional constraints on specified atoms.
    consref : str, default=None
        Specify a PDB file giving reference positions for constrained atoms.
    constraintscaling : float, default=None
        The harmonic constraint energy function is multiplied by this parameter.
    berendsenpressure : ('on', 'off'), str, default=None
        Set to enable the Berendsen pressure bath barostatic control.
    berendsenpressuretarget : float, default=None
        The target pressure (Bar) for the barostat.
    berendsenpressurerelaxationtime : float, default=None
        Relaxation time for the barostat (fs).
    tclforces : ('on', 'off'), str, default=None
        Enable TCL force scripting.
    minimize : int, default=None
        Number of steps of conjugate-gradient minimisation to perform.
    run : str, default=None
        The number of simulation iterations to perform.
    celldimension : str, default=None
        Dimensions of the unit cell.
    useconstantratio : ('on', 'off'), str, default=None
        Keep the ratio of the X-Y dimensions constant while allowing Z to fluctuate independently.
    amber : ('on', 'off'), str, default=None
        Indicate that the Amber force field is to be used.
    dielectric : float, default=None
        Dielectric constant.
    pairlistdist : str, default=None
        Specify the buffer size for grid cells.
    TCL : str, default=None
        Extra TCL code to be prepended to the input file

    Files
    -----
    Use these parameters to override the default search paths for these files.

    bincoordinates : str, default=None
        Filename for initial structure coordinates, in NAMD Bincoor format.
    binvelocities : str, default=None
        Initial velocity field, in NAMD Bincoor format.
    binindex : str, default=None
        Filename for index file to set initial timestep (as made by a check-point)
    structure : str, default=None
        CHARMM structure topology in PSF format
    parameters : str, default=None
        CHARMM force-field parameter file (PRM)
    extendedsystem : str, default=None
        If set, specifies an extended system .xsc file, from which a cell dimension will be read.
    coordinates : str, default=None
        Filename for initial structure coordinates, in PDB format.
    velocities : str, default=None
        Initial velocity field, in PDB format.
    parmfile : str, default=None
        The filename of the Amber PRMTOP parameter file.
    """
    def __init__(self):
        super().__init__(version=2)

        logger.warning('This class is for legacy support of older versions of ACEMD. '
                       'We recommend updating ACEMD to the latest version and using the Acemd class.')
        # ACEMD2 Options
        self._arg('temperature', 'float', 'Temperature of the thermostat in Kelvin.', None, val.Number(float, '0POS'))
        self._arg('restart', 'str', 'Restart simulation.', None, val.String(), valid_values=('on', 'off'))
        self._arg('restartfreq', 'int', 'Restart file frequency.', None, val.Number(int, '0POS'))
        self._arg('outputname', 'str', 'Output file name.', None, val.String())
        self._arg('xtcfile', 'str', 'Output XTC file name.', None, val.String())
        self._arg('xtcfreq', 'int', 'XTC sampling frequency in steps.', None, val.Number(int, '0POS'))
        self._arg('timestep', 'float', 'Simulation timestep.', None, val.Number(float, '0POS'))
        self._arg('rigidbonds', 'str', 'Enable holonomic constraints on all hydrogen-heavy atom bond terms', None,
                  val.String(), valid_values=('none', 'all'))
        self._arg('hydrogenscale', 'float', 'Amount by which to scale the mass of H atoms', None,
                  val.Number(float, '0POS'))
        self._arg('switching', 'str', 'Enable to apply smoothing and switching functions to electrostatic and VdW '
                                      'forces.', None, val.String(), valid_values=('on', 'off'))
        self._arg('switchdist', 'float', 'Range beyond which to begin to apply the switching functions.', None,
                  val.Number(float, '0POS'))
        self._arg('cutoff', 'str', 'Cutoff distance for direct-space electrostatic interaction evaluation.', None,
                  val.Number(float, '0POS'))
        self._arg('exclude', 'str', 'Which pairs of bonded atoms should be excluded from non-bonded interactions',
                  None, val.String(), valid_values=('none', '1-2', '1-3', '1-4', 'scaled1-4'))
        self._arg('scaling14', 'float', 'Scaling factor for 1-4 electrostatic interations.', None,
                  val.Number(float, '0POS'))
        self._arg('langevin', 'str', 'Enable the Langevin thermostat.', None, val.String(), valid_values=('on', 'off'))
        self._arg('langevintemp', 'float', 'The set point in K for the Langevin thermostat.', None,
                  val.Number(float, '0POS'))
        self._arg('langevindamping', 'float', 'Langevin damping constant gamma (1/ps)', None, val.Number(float, '0POS'))
        self._arg('pme', 'str', 'Enable the use of PME for long-range electrostatics.', None, val.String(),
                  valid_values=('on', 'off'))
        self._arg('pmegridspacing', 'float', 'The spacing of the PME mesh in 1/A.', None, val.Number(float, '0POS'))
        self._arg('fullelectfrequency', 'int', 'The frequency in interations between successive calculations of '
                                               'long-range (PME) electrostatics.', None, val.Number(int, '0POS'))
        self._arg('energyfreq', 'int', 'The frequency with which ACEMD will calculate system energies.', None,
                  val.Number(int, '0POS'))
        self._arg('constraints', 'str', 'Set to enable positional constraints on specified atoms.', None, val.String(),
                  valid_values=('on', 'off'))
        self._arg('consref', 'str', 'Specify a PDB file giving reference positions for constrained atoms.', None,
                  val.String())
        self._arg('constraintscaling', 'float', 'The harmonic constraint energy function is multiplied by this '
                                                'parameter.', None, val.Number(float, 'ANY'))
        self._arg('berendsenpressure', 'str', 'Set to enable the Berendsen pressure bath barostatic control.', None,
                  val.String(), valid_values=('on', 'off'))
        self._arg('berendsenpressuretarget', 'float', 'The target pressure (Bar) for the barostat.', None,
                  val.Number(float, '0POS'))
        self._arg('berendsenpressurerelaxationtime', 'float', 'Relaxation time for the barostat (fs).', None,
                  val.Number(float, '0POS'))
        self._arg('tclforces', 'str', 'Enable TCL force scripting.', None, val.String(), valid_values=('on', 'off'))
        self._arg('minimize', 'int', 'Number of steps of conjugate-gradient minimisation to perform.', None,
                  val.Number(int, '0POS'))
        self._arg('run', 'str', 'The number of simulation iterations to perform.', None)
        self._arg('celldimension', 'str', 'Dimensions of the unit cell.', None, val.Number(float, 'ANY'), nargs=3)
        self._arg('useconstantratio', 'str', 'Keep the ratio of the X-Y dimensions constant while allowing Z to '
                                             'fluctuate independently.', None, val.String(), valid_values=('on', 'off'))
        self._arg('amber', 'str', 'Indicate that the Amber force field is to be used.', None, val.String(),
                  valid_values=('on', 'off'))
        self._arg('dielectric', 'float', 'Dielectric constant.', None, val.Number(float, 'ANY'))
        self._arg('pairlistdist', 'str', 'Specify the buffer size for grid cells.', None, val.Number(float, 'ANY'))
        self._arg('TCL', 'str', 'Extra TCL code to be prepended to the input file', None, val.String(), nargs='*')

        # Files
        self._arg('bincoordinates', 'str', 'Filename for initial structure coordinates, in NAMD Bincoor format.', None,
                  val.String())
        self._arg('binvelocities', 'str', 'Initial velocity field, in NAMD Bincoor format.', None, val.String())
        self._arg('binindex', 'str', 'Filename for index file to set initial timestep (as made by a check-point)', None,
                  val.String())
        self._arg('structure', 'str', 'CHARMM structure topology in PSF format', None, val.String())
        self._arg('parameters', 'str', 'CHARMM force-field parameter file (PRM)', None, val.String())
        self._arg('extendedsystem', 'str', 'If set, specifies an extended system .xsc file, from which a cell dimension'
                                           ' will be read.', None, val.String())
        self._arg('coordinates', 'str', 'Filename for initial structure coordinates, in PDB format.', None,
                  val.String())
        self._arg('velocities', 'str', 'Initial velocity field, in PDB format.', None, val.String())
        self._arg('parmfile', 'str', 'The filename of the Amber PRMTOP parameter file.', None, val.String())


class Acemd(_Acemd):
    """ Class for configuring an ACEMD run.

    Note: default=None means the parameter is not set, which means that ACEMD will use it's own internal default. Check
    ACEMD documentation for its own defaults.

    Parameters
    ----------

    temperature : float, default=None
        Temperature of the thermostat in Kelvin.
    restart : str, default=None
        Restart simulation.
    trajectoryfile : str, default=None
        Output file name.
    trajectoryfreq : int, default=None
        Trajectory sampling frequency in steps.
    timestep : int, default=None
        Simulation timestep.
    pme : str, default=None
        Particle-mesh Ewald summation.
    switching : str, default=None
        Apply switching function to the van der Waals potential.
    switchdist : float, default=None
        Distance in Angstrom at which to begin applying the switching function.
    cutoff : float, default=None
        Real-space cutoff in Angstroms for electrostatics and van der Waals.
    thermostat : str, default=None
        Enable thermostatic control
    thermostattemp : float, default=None
        Target temperature (K) for thermostatic control
    thermostatdamping : float, default=None
        Damping constant for the Langevin thermostat in ps^-1
    restraints : str, default=None
        Restraining potentials
    barostat : str, default=None
        Enable pressure control
    barostatpressure : float, default=None
        The target pressure in bar
    useflexiblecell : str, default=None
        Allow X, Y and Z unit cell dimensions to vary independently
    useconstantarea : str, default=None
        Constrain the X,Y dimensions of the unit cell. Allow Z to vary independently.
    useconstantratio : str, default=None
        Constrain the X:Y ratio of the unit cell dimensions. Allow Z to vary independently.
    minimize : int, default=None
        The number of energy minimization steps to perform before commencing dynamics.
    run : str, default=None
        The length of simulation ro run. May be specified as a number of steps or as a time if one of the suffices "us", "ns", "ps", "fs" is used.
    celldimension : str, default=None
        The dimensions of the unit cell in Angstrom. Note that the unit cell must be cuboid. Overrides any dimension given in the "coordinates" PDB.
    implicit : str, default=None
        Set to True to enable implicit solvent simulations in AMBER.

    Files
    -----
    Use these parameters to override the default search paths for these files.

    bincoordinates : str, default=None
        Optional initial system geometry in NAMD BINCOOR format. If specified, overrides "coordinates"
    binvelocities : str, default=None
        Optional initial system velocity field in NAMD BINCOOR format. If specified, overrides field generated by "temperature" and "velocities"
    structure : str, default=None
        The filename of a CHARMM PSF file
    parameters : str, default=None
        The filename of a CHARMM PAR file
    parmfile : str, default=None
        The filename of an Amber PRMTOP file
    extendedsystem : str, default=None
        Filename of a NAMD XSC format file giving the periodic cell dimensions. Overrides "celldimension" and any dimensions in the "coordinates" PDB
    coordinates : str, default=None
        Mandatory initial system geometry in PDB format
    velocities : str, default=None
        Optional initial system velocity field in NAMD BINCOOR format. If specified, overrides field generated by "temperature"

    Examples
    --------
    >>> acemd = Acemd()
    Or you can load preset configurations for ACEMD which search for default files and use recommended settings.
    For example, to load the configuration for an equilibration run
    >>> equil = Acemd('equilibration')  # Loads the Acemd configuration for equilibration runs
    Now we write out the files needed for a run giving as input a folder containing the built structure files
    >>> equil.write('./build/', './equil/')
    Or to load the confifuration for a production run
    >>> prod = Acemd('production')
    >>> prod.write('./equil/', './prod/')
    """
    def __init__(self, config=None):
        super().__init__(version=3)
        # ACEMD3 Options
        self._cmdDeprecated('trajectoryfreq', 'trajectoryperiod')
        self._cmdDeprecated('switchdist', 'switchdistance')
        self._cmdDeprecated('thermostattemp', 'thermostattemperature')
        self._cmdDeprecated('useflexiblecell', 'barostatanisotropic')
        self._cmdDeprecated('useconstantarea', 'barostatconstxy')
        self._cmdDeprecated('useconstantratio', 'barostatconstratio')
        self._cmdDeprecated('celldimension', 'boxsize')

        self._arg('temperature', 'float', 'Temperature of the thermostat in Kelvin.', None, val.Number(float, 'ANY'))
        self._arg('restart', 'str', 'Restart simulation.', None, val.String())
        self._arg('trajectoryfile', 'str', 'Output file name.', None, val.String())
        self._arg('trajectoryperiod', 'int', 'Trajectory sampling frequency in steps.', None, val.Number(int, 'POS'))
        self._arg('timestep', 'int', 'Simulation timestep.', None, val.Number(int, 'POS'))
        self._arg('pme', 'str', 'Particle-mesh Ewald summation.', None, val.String())
        self._arg('switching', 'str', 'Apply switching function to the van der Waals potential.', None, val.String())
        self._arg('switchdistance', 'float', 'Distance in Angstrom at which to begin applying the switching function.',
                  None, val.Number(float, '0POS'))
        self._arg('cutoff', 'float', 'Real-space cutoff in Angstroms for electrostatics and van der Waals.', None,
                  val.Number(float, '0POS'))
        self._arg('thermostat', 'str', 'Enable thermostatic control', None, val.String())
        self._arg('thermostattemperature', 'float', 'Target temperature (K) for thermostatic control', None,
                  val.Number(float, '0POS'))
        self._arg('thermostatdamping', 'float', 'Damping constant for the Langevin thermostat in ps^-1', None,
                  val.Number(float, '0POS'))
        self._arg('restraints', 'str', 'Restraining potentials', None, val.Object(_Restraint), nargs='*')
        self._arg('barostat', 'str', 'Enable pressure control', None, val.String())
        self._arg('barostatpressure', 'float', 'The target pressure in bar', None, val.Number(float, '0POS'))
        self._arg('barostatanisotropic', 'str', 'Allow X, Y and Z unit cell dimensions to vary independently', None,
                  val.String())
        self._arg('barostatconstxy', 'str', 'Constrain the X,Y dimensions of the unit cell. Allow Z to vary '
                                            'independently.', None, val.String())
        self._arg('barostatconstratio', 'str', 'Constrain the X:Y ratio of the unit cell dimensions. Allow Z to vary '
                                             'independently.', None, val.String())
        self._arg('minimize', 'int', 'The number of energy minimization steps to perform before commencing dynamics.',
                  None, val.Number(int, '0POS'))
        self._arg('run', 'str', 'The length of simulation to run. May be specified as a number of steps or as a time '
                                'if one of the suffices "us", "ns", "ps", "fs" is used.', None, val.String())
        self._arg('boxsize', 'str', 'The dimensions of the unit cell in Angstrom. Note that the unit cell must '
                                          'be cuboid. Overrides any dimension given in the "coordinates" PDB.', None,
                  val.String())
        self._arg('implicit', 'str', 'Set to True to enable implicit solvent simulations in AMBER.', None, val.String())

        # Files
        self._arg('bincoordinates', 'str', 'Optional initial system geometry in NAMD BINCOOR format. If specified, '
                                           'overrides "coordinates"', None, val.String(), nargs='*')
        self._arg('binvelocities', 'str', 'Optional initial system velocity field in NAMD BINCOOR format. If '
                                          'specified, overrides field generated by "temperature" and "velocities"',
                  None, val.String(), nargs='*')
        self._arg('structure', 'str', 'The filename of a CHARMM PSF file', None, val.String(), nargs='*')
        self._arg('parameters', 'str', 'The filename of a CHARMM PAR file', None, val.String(), nargs='*')
        self._arg('parmfile', 'str', 'The filename of an Amber PRMTOP file', None, val.String(), nargs='*')
        self._arg('extendedsystem', 'str', 'Filename of a NAMD XSC format file giving the periodic cell dimensions. '
                                           'Overrides "boxsize" and any dimensions in the "coordinates" PDB',
                  None, val.String(), nargs='*')
        self._arg('coordinates', 'str', 'Mandatory initial system geometry in PDB format', None, val.String(),
                  nargs='*')
        self._arg('velocities', 'str', 'Optional initial system velocity field in NAMD BINCOOR format. If specified, '
                                       'overrides field generated by "temperature"', None, val.String(), nargs='*')

        if config is not None:
            self.readConfig(config)

    def readConfig(self, configfile):
        import json

        if not os.path.exists(configfile):
            from htmd.home import home
            configfile = os.path.join(home(shareDir=True), 'mdengine', 'acemd', 'config', '{}.json'.format(configfile))

        with open(configfile, 'r') as f:
            config = json.load(f)

        for key in config:
            if key == 'restraints':
                self.restraints = []
                for restr in ensurelist(config[key]['value']):
                    self.restraints.append(_Restraint._fromDict(restr))
            else:
                setattr(self, key, config[key]['value'])

    def write(self, inputdir='.', outputdir='run', overwrite=False):
        # if self.adaptive:
        #     self.binvelocities = None

        # if self.celldimension is None and self.extendedsystem is None:
        #     inmol = Molecule(os.path.join(inputdir, self.acemd.coordinates))
        #     coords = inmol.get('coords', sel='water')
        #     if coords.size == 0:  # It's a vacuum simulation
        #         coords = inmol.get('coords', sel='all')
        #         dim = np.max(coords, axis=0) - np.min(coords, axis=0)
        #         dim = dim + 12.
        #     else:
        #         dim = np.max(coords, axis=0) - np.min(coords, axis=0)
        #     self.celldimension = '{} {} {}'.format(dim[0], dim[1], dim[2])
        super().setup(inputdir, outputdir, overwrite)


class _TestAcemd(TestCase):

    def test_acemd2(self):
        from htmd.home import home
        from htmd.util import tempname
        import filecmp
        from glob import glob

        tmpdir = tempname()

        acemd = Acemd2()
        acemd.temperature = 350
        acemd.restart = 'on'
        acemd.restartfreq = 25
        acemd.outputname = 'output'
        acemd.xtcfile = 'myout.xtc'
        acemd.xtcfreq = 500
        acemd.timestep = 4
        acemd.rigidbonds = 'all'
        acemd.hydrogenscale = 0.3
        acemd.switching = 'on'
        acemd.switchdist = 9
        acemd.cutoff = 5
        acemd.exclude = 'scaled1-4'
        acemd.scaling14 = 5.6
        acemd.langevin = 'on'
        acemd.langevintemp = 300
        acemd.langevindamping = 0.8
        acemd.pme = 'on'
        acemd.pmegridspacing = 5
        acemd.fullelectfrequency = 3
        acemd.energyfreq = 10
        acemd.constraints = 'on'
        acemd.consref = '5dhfr_cube.pdb'
        acemd.constraintscaling = 3.2
        acemd.berendsenpressure = 'on'
        acemd.berendsenpressuretarget = 14
        acemd.berendsenpressurerelaxationtime = 2
        acemd.tclforces = 'on'
        acemd.minimize = 150
        acemd.run = 555
        acemd.celldimension = [3, 56, 2]
        acemd.useconstantratio = 'on'
        acemd.amber = 'off'
        acemd.dielectric = 16
        acemd.pairlistdist = 9
        acemd.TCL = 'on'
        acemd.bincoordinates = None
        acemd.binvelocities = None
        acemd.binindex = None
        acemd.structure = '5dhfr_cube.psf'
        acemd.parameters = 'par_all22_prot.inp'
        acemd.extendedsystem = None
        acemd.coordinates = '5dhfr_cube.pdb'
        acemd.velocities = None
        acemd.parmfile = None

        acemd.coordinates = '5dhfr_cube.pdb'
        acemd.setup(home(dataDir='dhfr/'), tmpdir, overwrite=True)

        # Compare with reference
        refdir = home(dataDir='test-acemd-v2')
        files = [os.path.basename(f) for f in glob(os.path.join(refdir, '*'))]
        match, mismatch, error = filecmp.cmpfiles(refdir, tmpdir, files, shallow=False)

        if len(mismatch) != 0 or len(error) != 0 or len(match) != len(files):
            raise RuntimeError('Different results produced by Acemd.write between {} and {} '
                               'in files {}.'.format(refdir, tmpdir, mismatch))

        import shutil
        shutil.rmtree(tmpdir)

    def test_acemd(self):
        import htmd.home
        homedir = htmd.home.home(dataDir='dhfr')

        r = list()
        r.append(GroupRestraint('resname MOL', 5, [(10, '10ns'), (5, '15ns'), (0, '20ns')], axes='z'))
        r.append(GroupRestraint('resname MOL', 5, [(10, '10ns'), (5, '15ns'), (0, '20ns')], axes='z',
                                fbcentre=[4, 2, 7.3]))
        r.append(GroupRestraint('resname MOL', 5, [(10, '10ns'), (5, '15ns'), (0, '20ns')], axes='z',
                                fbcentresel='protein'))
        r.append(AtomRestraint('name CA', 0.1, [(10, '10ns'), (5, '15ns'), (0, '20ns')]))
        r.append(AtomRestraint('name CA', [0.1, 0.5, 3], [(10, '10ns'), (5, '15ns'), (0, '20ns')]))

        acemd = Acemd()
        acemd.structure = '5dhfr_cube.psf'
        acemd.parameters = 'par_all22_prot.inp'
        acemd.coordinates = '5dhfr_cube.pdb'
        acemd.restraints = r
        acemd.temperature = 300
        acemd.run = '1000'
        acemd.setup(homedir, '/tmp/testdir', overwrite=True)

        print(acemd)

        expected_result = """
#
coordinates             structure.pdb
parameters              parameters
structure               structure.psf
temperature             300 
thermostattemperature   300
groupRestraint          "resname MOL" axes z width "5" setpoints 10@10ns 5@15ns 0@20ns
groupRestraint          "resname MOL" fbcentre "4 2 7.3" axes z width "5" setpoints 10@10ns 5@15ns 0@20ns
groupRestraint          "resname MOL" fbcentresel "protein" axes z width "5" setpoints 10@10ns 5@15ns 0@20ns
atomRestraint           "name CA" axes xyz width "0.1" setpoints 10@10ns 5@15ns 0@20ns
atomRestraint           "name CA" axes xyz width "0.1 0.5 3" setpoints 10@10ns 5@15ns 0@20ns
run                     1000
"""
        with open('/tmp/testdir/input', 'r') as f:
            lines = '\n'.join(f.readlines())

        import re
        lines = re.sub('\s+', ' ', lines)
        expected_result = re.sub('\s+', ' ', expected_result)

        self.assertTrue(expected_result.strip() == lines.strip(),
                        'Expected:\n{}\nGot:\n{}'.format(expected_result.strip(), lines.strip()))

    def test_production(self):
        from htmd.home import home
        import filecmp
        from htmd.util import tempname
        from glob import glob

        tmpdir = tempname()
        pdbid = '3PTB'

        prod = Acemd('production')
        prod.run = '2000'
        prod.trajectoryperiod = 200
        prod.temperature = 300
        prod.write(home(dataDir=os.path.join('test-acemd', pdbid, 'equil_out')), tmpdir)
        print(tmpdir)
        # Compare with reference
        refdir = home(dataDir=os.path.join('test-acemd', pdbid, 'prod'))
        files = [os.path.basename(f) for f in glob(os.path.join(refdir, '*'))]
        match, mismatch, error = filecmp.cmpfiles(refdir, tmpdir, files, shallow=False)

        if len(mismatch) != 0 or len(error) != 0 or len(match) != len(files):
            raise RuntimeError('Different results produced by Acemd production for '
                               'test {} between {} and {} in files {}.'.format(pdbid, refdir, tmpdir, mismatch))

    def test_equilibration(self):
        from htmd.home import home
        import filecmp
        from htmd.util import tempname
        from glob import glob
        from moleculekit.molecule import Molecule

        tmpdir = tempname()
        pdbid = '3PTB'
        builddir = home(dataDir=os.path.join('test-acemd', pdbid, 'build'))

        equil = Acemd('equilibration')
        mol = Molecule(os.path.join(builddir, 'structure.pdb'))
        celldim = mol.coords.max(axis=0) - mol.coords.min(axis=0)
        equil.boxsize = ' '.join(['{:3.1f}'.format(val) for val in celldim.squeeze()])
        equil.run = '2000'
        equil.trajectoryperiod = 200
        equil.temperature = 300

        equil.write(builddir, tmpdir)

        # Compare with reference
        refdir = home(dataDir=os.path.join('test-acemd', pdbid, 'equil'))
        files = [os.path.basename(f) for f in glob(os.path.join(refdir, '*'))]
        match, mismatch, error = filecmp.cmpfiles(refdir, tmpdir, files, shallow=False)

        if len(mismatch) != 0 or len(error) != 0 or len(match) != len(files):
            raise RuntimeError('Different results produced by Acemd equilibration for '
                               'test {} between {} and {} in files {}.'.format(pdbid, refdir, tmpdir, mismatch))


if __name__ == "__main__":
    import unittest
    unittest.main(verbosity=2)




# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import warnings
import numpy as np
import random
from htmd.projections.metric import Metric, _projectionGenerator
from htmd.units import convert as unitconvert
from joblib import Parallel, delayed
import logging
logger = logging.getLogger(__name__)


class TICA(object):
    """ Class for calculating the TICA projections of a MetricData  object

    Time-based Independent Component Analysis
    Projects your data on the slowest coordinates identified for a
    given lagtime.

    Parameters
    ----------
    data : :class:`MetricData <htmd.metricdata.MetricData>` object
        The object whose data we wish to project onto the top TICA dimensions
    lag : int
        The correlation lagtime to use for TICA
    units : str
        The units of lag. Can be 'frames' or any time unit given as a string.
    dimensions : list
        A list of dimensions of the original data on which to apply TICA. All other dimensions will stay unaltered.
        If None is given, it will apply on all dimensions.
    njobs : int
        Number of jobs to spawn for parallel computation of TICA components. If None it will use the default from htmd.config.

    Example
    -------
    >>> from htmd.projections.tica import TICA
    >>> metr = Metric(sims)
    >>> metr.set(MetricSelfDistance('protein and name CA'))
    >>> data = metr.project()
    >>> tica = TICA(data, 20)
    >>> datatica = tica.project(3)
    Alternatively you can pass a Metric object to TICA. Uses less memory but is slower.
    >>> metr = Metric(sims)
    >>> metr.set(MetricSelfDistance('protein and name CA'))
    >>> slowtica = TICA(metr, 20)
    >>> datatica = slowtica.project(3)


    References
    ----------
    Perez-Hernandez, G. and Paul, F. and Giorgino, T. and de Fabritiis, G.
    and Noe, F. (2013) Identification of slow molecular order parameters
    for Markov model construction. J. Chem. Phys., 139 . 015102.
    """

    def __init__(self, data, lag, units='frames', dimensions=None, njobs=None):
        from pyemma.coordinates.transform.tica import TICA as TICApyemma
        from tqdm import tqdm
        from htmd.util import _getNjobs

        self.data = data
        self.dimensions = dimensions
        self.njobs = njobs if njobs is not None else _getNjobs()

        if isinstance(data, Metric):  # Memory efficient TICA projecting trajectories on the fly
            if units != 'frames':
                raise RuntimeError('Cannot use delayed projection TICA with units other than frames for now. Report this to HTMD issues.')
            self.tic = TICApyemma(lag)
            metr = data

            pbar = tqdm(total=len(metr.simulations))
            for proj in _projectionGenerator(metr, self.njobs):
                for pro in proj:
                    if pro is None:
                        continue
                    if self.dimensions is None:
                        self.tic.partial_fit(pro[0])
                    else:  # Sub-select dimensions for fitting
                        self.tic.partial_fit(pro[0][:, self.dimensions])
                pbar.update(len(proj))
            pbar.close()
        else:  # In-memory TICA
            lag = unitconvert(units, 'frames', lag, data.fstep)
            if lag == 0:
                raise RuntimeError('Lag time conversion resulted in 0 frames. Please use a larger lag-time for TICA.')

            self.tic = TICApyemma(lag)
            if self.dimensions is None:
                datalist = data.dat.tolist()
            else:  # Sub-select dimensions for fitting
                datalist = [x[:, self.dimensions].copy() for x in data.dat]
            self.tic.fit(datalist)

    def project(self, ndim=None):
        """ Projects the data object given to the constructor onto the top `ndim` TICA dimensions

        Parameters
        ----------
        ndim : int
            The number of TICA dimensions we want to project the data on. If None is given it will use choose a number
            of dimensions to cover 95% of the kinetic variance.

        Returns
        -------
        dataTica : :class:`MetricData <htmd.metricdata.MetricData>` object
            A new :class:`MetricData <htmd.metricdata.MetricData>` object containing the TICA projected data

        Example
        -------
        >>> from htmd.projections.tica import TICA
        >>> tica = TICA(data,20)
        >>> dataTica = tica.project(5)
        """
        from tqdm import tqdm
        if ndim is not None:
            self.tic.set_params(dim=ndim)

        keepdata = []
        keepdim = None
        keepdimdesc = None
        if isinstance(self.data, Metric):  # Memory efficient TICA projecting trajectories on the fly
            proj = []
            refs = []
            fstep = None

            metr = self.data
            k = -1
            droppedsims = []
            pbar = tqdm(total=len(metr.simulations))
            for projecteddata in _projectionGenerator(metr, self.njobs):
                for pro in projecteddata:
                    k += 1
                    if pro is None:
                        droppedsims.append(k)
                        continue
                    if self.dimensions is not None:
                        numDimensions = pro[0].shape[1]
                        keepdim = np.setdiff1d(range(numDimensions), self.dimensions)
                        keepdata.append(pro[0][:, keepdim])
                        proj.append(self.tic.transform(pro[0][:, self.dimensions]).astype(np.float32))  # Sub-select dimensions for projecting
                    else:
                        proj.append(self.tic.transform(pro[0]).astype(np.float32))
                    refs.append(pro[1])
                    if fstep is None:
                        fstep = pro[2]
                pbar.update(len(projecteddata))
            pbar.close()

            simlist = self.data.simulations
            simlist = np.delete(simlist, droppedsims)
            ref = np.array(refs, dtype=object)
            parent = None
            if self.dimensions is not None:
                from htmd.projections.metric import _singleMolfile
                from moleculekit.molecule import Molecule
                (single, molfile) = _singleMolfile(metr.simulations)
                if single:
                    keepdimdesc = metr.getMapping(Molecule(molfile))
                    keepdimdesc = keepdimdesc.iloc[keepdim]
        else:
            if ndim is not None and self.data.numDimensions < ndim:
                raise RuntimeError('TICA cannot increase the dimensionality of your data. Your data has {} dimensions and you requested {} TICA dimensions'.format(self.data.numDimensions, ndim))

            if self.dimensions is not None:
                keepdim = np.setdiff1d(range(self.data.numDimensions), self.dimensions)
                keepdata = [x[:, keepdim] for x in self.data.dat]
                if self.data.description is not None:
                    keepdimdesc = self.data.description.iloc[keepdim]
            proj = self.tic.get_output()
            simlist = self.data.simlist
            ref = self.data.ref
            fstep = self.data.fstep
            parent = self.data

        # If TICA is done on a subset of dimensions, combine non-projected data with projected data
        if self.dimensions is not None:
            newproj = []
            for k, t in zip(keepdata, proj):
                newproj.append(np.hstack((k, t)))
            proj = newproj

        if ndim is None:
            ndim = self.tic.dimension()
            logger.info('Kept {} dimension(s) to cover 95% of kinetic variance.'.format(ndim))

        from htmd.metricdata import MetricData
        datatica = MetricData(dat=np.array(proj), simlist=simlist, ref=ref, fstep=fstep, parent=parent)
        from pandas import DataFrame
        # TODO: Make this messy pandas creation cleaner. I'm sure I can append rows to DataFrame
        types = []
        indexes = []
        description = []
        for i in range(ndim):
            types += ['tica']
            indexes += [-1]
            description += ['TICA dimension {}'.format(i+1)]
        datatica.description = DataFrame({'type': types, 'atomIndexes': indexes, 'description': description})

        if self.dimensions is not None and keepdimdesc is not None:  # If TICA is done on a subset of dims
            datatica.description = keepdimdesc.append(datatica.description, ignore_index=True)

        return datatica


if __name__ == '__main__':
    from htmd.simlist import simlist
    from glob import glob
    from moleculekit.projections.metricdistance import MetricSelfDistance
    from htmd.home import home
    from os.path import join

    testfolder = home(dataDir='villin')

    sims = simlist(glob(join(testfolder, '*', '')), join(testfolder, 'filtered.pdb'))
    met = Metric(sims[0:2])
    met.set(MetricSelfDistance('protein and name CA'))
    data = met.project()
    data.fstep = 0.1

    tica = TICA(data, 2, dimensions=range(2, 10))
    datatica = tica.project(2)
    tica5 = TICA(data, 0.2, units='ns', dimensions=range(2, 10))
    datatica5 = tica5.project(2)
    expected = [[ 3.69098878, -0.33862674,  0.85779184],
                [ 3.77816105, -0.31887317,  0.87724227],
                [ 3.83537507, -0.11878026,  0.65236956]]
    assert np.allclose(np.abs(datatica.trajectories[0].projection[-3:, -3:]), np.abs(np.array(expected, dtype=np.float32)), rtol=0, atol=0.01)
    assert np.allclose(np.abs(datatica5.trajectories[0].projection[-3:, -3:]), np.abs(np.array(expected, dtype=np.float32)), rtol=0, atol=0.01)
    assert np.all(datatica.description.iloc[[587, 588]].type == 'tica')
    assert np.all(datatica.description.iloc[range(587)].type == 'distance')
    print('In-memory TICA with subset of dimensions passed test.')

    tica2 = TICA(met, 2, dimensions=range(2, 10))
    datatica2 = tica2.project(2)
    assert np.allclose(np.abs(datatica2.trajectories[0].projection[-3:, -3:]), np.abs(np.array(expected, dtype=np.float32)), rtol=0, atol=0.01)
    assert np.all(datatica2.description.iloc[[587, 588]].type == 'tica')
    assert np.all(datatica2.description.iloc[range(587)].type == 'distance')
    print('Streaming TICA with subset of dimensions passed test.')

    #assert np.max(np.abs(datatica.dat[0][:, -2:]) - np.abs(datatica2.dat[0][:, -2:])) < 0.01, 'Streaming and memory subdim TICA inconsistent.'

    tica3 = TICA(data, 2)
    datatica3 = tica3.project(2)
    expected = [[-1.36328638, -0.35354128],
                [-1.35348749, -0.13028328],
                [-1.43249917, -0.31004715]]
    assert np.allclose(np.abs(datatica3.trajectories[0].projection[-3:, :]), np.abs(np.array(expected, dtype=np.float32)), rtol=0, atol=0.01)
    assert np.all(datatica3.description.iloc[[0, 1]].type == 'tica')
    print('In-memory TICA passed test.')

    tica4 = TICA(met, 2)
    datatica4 = tica4.project(2)
    assert np.allclose(np.abs(datatica4.trajectories[0].projection[-3:, :]), np.abs(np.array(expected, dtype=np.float32)), rtol=0, atol=0.01)
    assert np.all(datatica4.description.iloc[[0, 1]].type == 'tica')
    print('Streaming TICA passed test.')

    assert np.max(np.abs(datatica4.trajectories[0].projection) - np.abs(datatica3.trajectories[0].projection)) < 0.01, 'Streaming and memory TICA inconsistent.'



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import numpy as np
from tempfile import NamedTemporaryFile
from os import path
from htmd.home import home
import os
import shutil
from subprocess import call, check_output
from htmd.util import tempname
from moleculekit.molecule import Molecule
from glob import glob
import logging
logger = logging.getLogger(__name__)


def dock(protein, ligand, center=None, extent=None, numposes=20, babelexe='obabel', vinaexe=None):
    """ Molecular docking, using Vina

    If centre and extent are not provided, docking will be performed over the whole protein

    Parameters
    ----------
    protein : :class:`Molecule <moleculekit.molecule.Molecule>` object
        Molecule object representing the receptor
    ligand : :class:`Molecule <moleculekit.molecule.Molecule>` object
        Molecule object representing the ligand to dock
    center : list
        3-vec centre of of the search bounding box (optional)
    extent : list
        3-vec linear extent of the search bounding box (optional)
    numposes : int
        Number of poses to return. Vina cannot return more than 20 poses.
    babelexe : str
        Path to babel executable.
    vinaexe : str
        Path to AutoDock Vina executable.

    Returns
    -------
    poses
        Molecule object representing the N<10 best poses
    scores
        3x num_poses matrix containing kcal, rmsd lb, rmsd ub

    Examples
    --------
    >>> poses, scoring = dock(protein, ligand)
    >>> poses, scoring = dock(protein, ligand, center=[ 10., 5., 12. ], extent=[ 15., 15., 15. ] )

    """
    if np.size(protein.coords, 2) != 1 or np.size(ligand.coords, 2) != 1:
        raise NameError('Protein and ligand Molecules should be single frames')

    buffer = 10.  # Angstrom buffer around protein for whole-protein docking
    c_min = np.min(protein.coords, 0).reshape((1, 3))[0]
    c_max = np.max(protein.coords, 0).reshape((1, 3))[0]

    if center is None:
        center = (buffer + (c_max + c_min)) / 2
    if extent is None:
        extent = (c_max - c_min) + buffer

    # babel -i pdb protein.pdb  -o pdbqt protein.pdbqt -xr
    # babel -i pdb ligand.pdb   -o pdbqt ligand.pdbqt -xhn
    # vina --ligand ligand.pdbqt --receptor protein.pdbqt --center_x 0. --center_y 0. --center_z 0. --size_x 60. --size_y 60. --size_z 60 --exhaustiveness 10
    # babel -m -i pdbqt ligand_out.pdbqt -o pdb out_.pdb -xhn

    protein_pdb = tempname(suffix=".pdb")
    ligand_mol2 = tempname(suffix=".mol2")
    output_pdb = tempname(suffix="_.pdb")
    output_prefix = path.splitext(output_pdb)[0]

    protein_pdbqt = tempname(suffix=".pdbqt")
    ligand_pdbqt = tempname(suffix=".pdbqt")
    output_pdbqt = tempname(suffix=".pdbqt")

    protein.write(protein_pdb)
    lig2 = ligand.copy()
    lig2.atomtype = lig2.element  # babel does not understand mol2 atomtypes and requires elements instead
    lig2.write(ligand_mol2)

    # Dirty hack to remove the 'END' line from the PDBs since babel hates it
    with open(protein_pdb, 'r') as f:
        lines = f.readlines()
    with open(protein_pdb, 'w') as f:
        f.writelines(lines[:-1])
    # End of dirty hack

    try:
        if vinaexe is None:
            import platform
            suffix = ''
            if platform.system() == "Windows":
                suffix = '.exe'
            vinaexe = '{}-vina{}'.format(platform.system(), suffix)

        vinaexe = shutil.which(vinaexe, mode=os.X_OK)
        if not vinaexe:
            raise NameError('Could not find vina, or no execute permissions are given')
    except:
        raise NameError('Could not find vina, or no execute permissions are given')
    try:
        babelexe = shutil.which(babelexe, mode=os.X_OK)
        if babelexe is None:
            raise NameError('Could not find babel, or no execute permissions are given')
    except:
        raise NameError('Could not find babel, or no execute permissions are given')

    call([babelexe, '-i', 'pdb', protein_pdb, '-o', 'pdbqt', '-O', protein_pdbqt, '-xr'])
    if np.all(ligand.charge != 0):
        logger.info('Charges detected in ligand and will be used for docking.')
        call([babelexe, '-i', 'mol2', ligand_mol2, '-o', 'pdbqt', '-O', ligand_pdbqt, '-xn', '-xh'])
    else:
        logger.info('Charges were not defined for all atoms. Will guess charges anew using gasteiger method.')
        call([babelexe, '-i', 'mol2', ligand_mol2, '-o', 'pdbqt', '-O', ligand_pdbqt, '-xn', '-xh', '--partialcharge', 'gasteiger'])

    if not path.isfile(ligand_pdbqt):
        raise NameError('Ligand could not be converted to PDBQT')
    if not path.isfile(protein_pdbqt):
        raise NameError('Protein could not be converted to PDBQT')

    call([vinaexe, '--receptor', protein_pdbqt, '--ligand', ligand_pdbqt, '--out', output_pdbqt,
          '--center_x', str(center[0]), '--center_y', str(center[1]), '--center_z', str(center[2]),
          '--size_x', str(extent[0]), '--size_y', str(extent[1]), '--size_z', str(extent[2]), '--num_modes', str(numposes)])

    call([babelexe, '-m', '-i', 'pdbqt', output_pdbqt, '-o', 'pdb', '-O', output_pdb, '-xhn'])

    from natsort import natsorted
    outfiles = natsorted(glob('{}*.pdb'.format(output_prefix)))

    scoring = []
    poses = []
    for i, ligf in enumerate(outfiles):
        scoring.append(_parseScoring(ligf))
        l = Molecule(ligf)
        l.viewname = 'Pose {}'.format(i)
        poses.append(l)

    os.remove(protein_pdb)
    os.remove(ligand_mol2)
    os.remove(protein_pdbqt)
    os.remove(ligand_pdbqt)
    os.remove(output_pdbqt)

    return poses, np.array(scoring)


def _parseScoring(outf):
    kcal = rmsdlb = rmsdub = None
    with open(outf, 'r') as f:
        for line in f:
            if line.startswith('REMARK VINA RESULT:'):
                pieces = line.split()
                kcal = float(pieces[3])
                rmsdlb = float(pieces[4])
                rmsdub = float(pieces[5])
                break
    if kcal is None or rmsdlb is None or rmsdub is None:
        raise RuntimeError('Could not parse vina output correctly {}'.format(outf))
    return kcal, rmsdlb, rmsdub

if __name__ == "__main__":
    from moleculekit.molecule import Molecule
    from os import path
    from htmd.home import home
    protein = Molecule(path.join(home(), 'data', 'docking', 'protein.pdb'))
    ligand = Molecule(path.join(home(), 'data', 'docking', 'ligand.pdb'))
    #poses, scoring = dock(protein, ligand)




from jobqueues.slurmqueue import SlurmQueue
import logging


logger = logging.getLogger(__name__)

logger.warning('Please do not import from htmd.queues.slurmqueue. This will be deprecated. '
                'Use jobqueues.slurmqueue instead.')


# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#

# Author: Juan Eiros Zamora (je714@ic.ac.uk) & Gil Ferreira Hoben (gf712@ic.ac.uk) Organisation: Imperial College London
# Modified by: Stefan Doerr

from moleculekit.molecule import Molecule
from htmd.apps.pmemd import Pmemd
from htmd.protocols.oldprotocolinterface import ProtocolInterface, TYPE_INT, TYPE_FLOAT, RANGE_0POS, RANGE_POS, RANGE_ANY
import os
import numpy as np
import logging
import shutil
from htmd.decorators import _Deprecated
logger = logging.getLogger(__name__)


@_Deprecated('1.13.6')
class Equilibration(ProtocolInterface):

    """ Equilibration protocol

        Equilibration protocol for globular and membrane proteins
        It includes a flatbottom potential box to retrain a ligand
        for example within this box.

        Parameters
        ----------
        numsteps : int, default=0
            Number of steps to run the simulations in units of 4fs
        temperature : float, default=300
            Temperature of the thermostat in Kelvin
        k : float, default=0
            Force constant of the flatbottom potential in kcal/mol/A^2. E.g. 5
        reference : str, default='none'
            Reference selection to use as dynamic center of the flatbottom box.
        selection : str, default='none'
            Selection of atoms to apply the flatbottom potential
        box : list, default=[0, 0, 0, 0, 0, 0]
            Position of the flatbottom box in term of the reference center given as [xmin, xmax, ymin, ymax, zmin, zmax]
        useconstantratio : bool, default=False
            For membrane protein simulations set it to true so that the barostat does not modify the xy aspect ratio.
        constraints : dict, default={'protein and noh and not name CA': 0.1, 'protein and name CA': 1}
            A dictionary containing as keys the atomselections of the constraints and as values the constraint scaling factor. 0 factor means no constraint, 1 full constraints and in between values are used for scaling. The order with which the constraints are applied is random, so make atomselects mutually exclusive to be sure you get the correct constraints.

        Example
        -------
        >>> from htmd.protocols.equilibration_v1 import Equilibration
        >>> md = Equilibration()
        >>> md.numsteps = 10000000
        >>> md.temperature = 300
        >>> md.useconstantratio = True  # only for membrane sims
        >>> # this is only needed for setting the flatbottom potential, otherwise remove it
        >>> md.reference = 'protein and resid 293'
        >>> md.selection = 'segname L and noh'
        >>> md.box = [-25, 25, -25, 25, 43, 45]
        >>> md.k = 5
        >>> md.write('./build','./equil')
    """

    def __init__(self):
        super().__init__()
        self._cmdObject('amber', ':class:`MDEngine <htmd.apps.app.App>` object', 'MD engine', None, Pmemd)

        self.amber = Pmemd()
        self.amber.imin = 0
        self.amber.ntx = 1
        self.amber.irest = 0
        self.amber.ntc = 2
        self.amber.ntf = 2
        self.amber.tol = 0.0000001
        self.amber.nstlim = 25000
        self.amber.ntt = 3
        self.amber.gamma_ln = 1.0
        self.amber.ntr = 1
        self.amber.ig = -1
        self.amber.ntpr = 1000
        self.amber.ntwr = 10000
        self.amber.ntwx = 1000
        self.amber.dt = 0.004
        self.amber.nmropt = 1
        self.amber.ntb = 1
        self.amber.ntp = 0
        self.amber.cut = 8.0
        self.amber.ioutfm = 1
        #  Molecular dynamics (Manual section 18.6.6)

        self.amber.outputnc = 'Equilibration.nc'
        self.amber.FORTRAN = """ heating 100K
 &cntrl
   imin=IMIN, ntx=NTX, irest=IREST,
   ntc=NTC, ntf=NTF, tol=TOL,
   nstlim=NSTLIM, ntt=NTT, gamma_ln=GAMMA_LN,
   ntr=NTR, ig=IG,
   ntpr=NTPR, ntwr=NTWR, ntwx=NTWX,
   dt=DT, nmropt=NMROPT,
   ntb=NTB, ntp=NTP, cut=CUT, ioutfm=IOUTFM,
 / \n"""
        #self.amber.FORTRAN = ''' HEATING\n &cntrl\n'''
        self.amber.bash = '''ENGINE -O -i INPUT -o OUTPUT -p TOPOLOGY -c RESTART -x TRAJOUT -r RESTOUT'''  # FIXME: add -ref REFERENCE

    def _findFiles(self, inputdir):
        # Tries to find default files if the given don't exist
        defaults = {'coordinates': ('structure.rst', 'structure.inpcrd', 'structure.mdcrd'),
                    'parmfile': ('structure.prmtop',)}

        for field in defaults:
            userval = self.amber.__dict__[field]
            if userval is not None and not os.path.exists(os.path.join(inputdir, userval)):
                self.amber.__dict__[field] = None

            if self.amber.__dict__[field] is None:
                for val in defaults[field]:
                    if os.path.exists(os.path.join(inputdir, val)):
                        self.amber.__dict__[field] = val
                        break

            if userval is not None and self.amber.__dict__[field] is not None and self.amber.__dict__[field] != userval:
                logger.warning('Could not locate structure file {}. Using {} instead.'.format(
                    os.path.join(inputdir, userval), os.path.join(
                        inputdir, self.amber.__dict__[field])
                ))
            elif self.amber.__dict__[field] is None:
                raise RuntimeError('Could not locate any {f:} file in {i:}. '
                                   'Please set the Equilibration.acemd.{f:} property to '
                                   'point to the {f:} file'.format(f=field, i=inputdir))

    # FIXME: add inputdir and outputdir to write method
    # need to decide how to organize MD engine input files!!!
    # Probably will create a bash script here with all the information needed to run pmemd, e.g.:
    # pmemd.cuda_SPFP -O -i Equil.in -o Equil.out -c structure.rst -p structure \
    # -r EquilRestart.rst -x EquilRestart.nc (-ref Min.rst)

    def write(self, inputdir, outputdir):
        """ Write the equilibration protocol

        Writes the equilibration protocol and files into a folder for execution
        using files inside the inputdir directory

        Parameters
        ----------
        inputdir : str
            Path to a directory containing the files produced by a build process.
        outputdir : str
            Directory where to write the equilibration setup files.

        Examples
        --------
        >>> md = Equilibration()
        >>> md.write('./build','./equil')
        """
        # this code automatically loops through the amber attributes and assigns the given value to
        # the parameter.

        self._findFiles(inputdir)

        restartfile = os.path.join(inputdir, self.amber.coordinates)
        topology = os.path.join(inputdir, self.amber.parmfile)

        if self.amber.consref is not None:
            reference = os.path.join(inputdir, self.amber.consref)

        # FOR NOW WE WILL HAVE A FIXED STRING LIKE IN THE ACEMD PROTOCOL
        # THIS WAY WE CAN CONTROL BETTER WHAT PARAMETERS CAN BE SPECIFIED
        # IN THE PROTOCOL

        # protocol=[]
        # i=0
        # for key, value in self.amber.__dict__.items():

        #     if not key.startswith('_') and key not in ['bincoordinates', 'coordinates',
        #                                                'consref', 'parmfile','FORTRAN',
        #                                                'bash', 'outputnc']:

        #         # cleans up the .in file a bit
        #         if i % 3 == 0 and i>0:
        #             protocol.append('\n   {}={}'.format(key,value))
        #         else:
        #             protocol.append('{}={}'.format(key,value))

        #         i+=1

        # # format the FORTRAN file with MD parameters
        # self.amber.FORTRAN = self.amber.FORTRAN+'   '+(', '.join(protocol)) + "\n /"

        # preparing the FORTRAN file
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'IMIN', str(self.amber.imin))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'NTX', str(self.amber.ntx))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'IREST', str(self.amber.irest))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'NTC', str(self.amber.ntc))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'NTF', str(self.amber.ntf))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'TOL', '{:.10f}'.format(self.amber.tol))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'NSTLIM', str(self.amber.nstlim))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'NTT', str(self.amber.ntt))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'GAMMA_LN', str(self.amber.gamma_ln))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'NTR', str(self.amber.ntr))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'IG', str(self.amber.ig))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'NTPR', str(self.amber.ntpr))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'NTWR', str(self.amber.ntwr))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'NTWX', str(self.amber.ntwx))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'DT', str(self.amber.dt))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'NMROPT', str(self.amber.nmropt))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'NTB', str(self.amber.ntb))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'NTP', str(self.amber.ntp))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'CUT', str(self.amber.cut))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'IOUTFM', str(self.amber.ioutfm))

        # write out the FORTRAN file, for now let's call it Equilibration.in
        with open(os.path.join(outputdir, 'Equilibration.in'), 'w') as text_file:
            text_file.write(self.amber.FORTRAN)

        # prepare the bash file
        # putting these strings as placeholders for now
        self.amber.bash = self.amber.bash.replace('INPUT', 'Equilibration.in')
        self.amber.bash = self.amber.bash.replace(
            'OUTPUT', 'Equilibration.out')
        self.amber.bash = self.amber.bash.replace(
            'TOPOLOGY', self.amber.parmfile)
        self.amber.bash = self.amber.bash.replace(
            'RESTART', self.amber.coordinates)
        self.amber.bash = self.amber.bash.replace(
            'TRAJOUT', self.amber.outputnc)
        self.amber.bash = self.amber.bash.replace(
            'RESTOUT', 'Equilibration_new.rst')
        # FIXME: add ref
        #self.amber.bash = self.amber.bash.replace('REFERENCE', self.amber.reference)

        with open(os.path.join(outputdir, 'MD.sh'), 'w') as text_file:
            text_file.write(self.amber.bash)

        shutil.copy(restartfile, outputdir)
        shutil.copy(topology, outputdir)
        # FIXME: add ref
        #shutil.copy(consref, outputdir)



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#

# Author: Juan Eiros Zamora (je714@ic.ac.uk) & Gil Ferreira Hoben (gf712@ic.ac.uk) Organisation: Imperial College London
# Modified by: Stefan Doerr

# Code to setup simulations with Amber
# Description of parameters is mainly being copied from the AMBER 15 manual
# http://ambermd.org/doc12/Amber15.pdf
# However, some parameter options have different names from Amber to
# to keep protocol creation as consistent as possible with acemd

import os
import shutil
import numpy as np
from htmd.protocols.oldprotocolinterface import ProtocolInterface, TYPE_INT, TYPE_FLOAT, RANGE_0POS, RANGE_POS, RANGE_ANY
from htmd.decorators import _Deprecated


@_Deprecated('1.13.6')
class Pmemd(ProtocolInterface):
    _defaultfnames = {'bincoordinates': 'input.nc',
                      'parameters': 'parameters', 'coordinates': 'structure.rst',
                      'parmfile': 'structure.prmtop'}

    def __init__(self):
        super().__init__()

        self._files = {}

        # GENERAL MINIMIZATION AND DYNAMICS PARAMETERS (MANUAL SECTION 18.6)
        #  General flags describing the calculation (Manual section 18.6.1)
        self._cmdList(key='imin', datatype='int',
                      descr='Flag to run minimization.',
                      default=0, valid_values=[0, 1, 5])
        self._cmdList(key='nmropt', datatype='int',
                      descr='Flag to apply NMR restraints',
                      default=0, valid_values=[0, 1, 2])

        #  Nature and format of the input (Manual section 18.6.2)
        self._cmdList(key='ntx', datatype='int',
                      descr="""Option to read the initial coordinates,
                            velocities and box size from the inpcrd file.""",
                      default=2, valid_values=[1, 2, 4, 5, 6])
        self._cmdList(key='irest', datatype='int',
                      descr='Flag to restart a simulation.',
                      default=1, valid_values=[0, 1])

        #  Nature and format of the output (Manual section 18.6.3)
        self._cmdList(key='ntxo', datatype='int',
                      descr="""Format of the final coordinates, velocities,
                            and box size (if constant volume or pressure run)
                            written to file "restrt".""", default=2,
                      valid_values=[1, 2])
        self._cmdValue(key='ntpr', datatype='int', realdatatype=TYPE_INT,
                       descr="""Every ntpr steps, energy information will be
                             printed in human-readable form to files "mdout"
                             and "mdinfo".""", default=50, valid_range=RANGE_POS)
        self._cmdValue(key='ntave', datatype='int', realdatatype=TYPE_INT,
                       descr="""Every ntave steps of dynamics, running averages
                             of average energies and fluctuations over the last
                             ntave steps will be printed out.""", default=0,
                       valid_range=RANGE_0POS)
        self._cmdValue(key='ntwr', datatype='int', realdatatype=TYPE_INT,
                       descr="""Every ntwr steps during dynamics, the “restrt”
                             file will be written.""", default=None, valid_range=RANGE_POS)
        self._cmdList(key='iwrap', datatype='int',
                      descr="""If iwrap = 1, the coordinates written to the
                            restart and trajectory files will be "wrapped"
                            into a primary box. If iwrap = 0, no wrapping will
                            be performed.""", default=0, valid_values=[0, 1])
        self._cmdValue(key='ntwx', datatype='int', realdatatype=TYPE_INT,
                       descr="""Every ntwx steps, the coordinates will be
                             written to the mdcrd file. If ntwx = 0, no
                             coordinate trajectory file will be written.""",
                       default=0, valid_range=RANGE_0POS)
        self._cmdValue(key='ntwv', datatype='int', realdatatype=TYPE_INT,
                       descr="""Every ntwv steps, the velocities will be written
                             to the mdvel file. If ntwv = 0, no velocity
                             trajectory file will be written. If ntwv = -1,
                             velocities will be written to mdcrd, which then
                             becomes a combined coordinate/velocity trajectory
                             file, at the interval defined by ntwx.""",
                       default=0, valid_range=RANGE_ANY)
        self._cmdValue(key='ntwf', datatype='int', realdatatype=TYPE_INT,
                       descr="""Every ntwf steps, the forces will be written to
                             the mdfrc file. If ntwf = 0, no force trajectory
                             file will be written. If ntwf = -1, forces will be
                             written to the mdcrd, which then becomes a combined
                             coordinate/force trajectory file, at the interval
                             defined by ntwx.""", default=0, valid_range=RANGE_ANY)
        self._cmdValue(key='ntwe', datatype='int', realdatatype=TYPE_INT,
                       descr="""Every ntwe steps, the energies and temperatures
                             will be written to file "mden" in a compact form.
                             If ntwe = 0 then no mden file will be written.""",
                       default=0, valid_range=RANGE_0POS)
        self._cmdList(key='ioutfm', datatype='int',
                      descr="""The format of coordinate and velocity trajectory
                            files (mdcrd, mdvel and inptraj). 0 is ASCII and 1
                            is Binary NetCDF""", default=1, valid_values=[0, 1])
        self._cmdValue(key='ntwprt', datatype='int', realdatatype=TYPE_INT,
                       descr="""The number of atoms to include in trajectory
                             files (mdcrd and mdvel). If ntwprt = 0, all atoms
                             will be included.""", default=0, valid_range=RANGE_0POS)
        self._cmdList(key='idecomp', datatype='int',
                      descr="""Perform energy decomposition according to a
                            chosen scheme.""", default=0, valid_values=[0, 1, 2, 3, 4])

        #  Frozen or restrained atoms (Manual section 18.6.4)
        self._cmdList(key='ibelly', datatype='int',
                      descr="""Flag for belly type dynamics. If set to 1, a
                            subset of the atoms in the system will be allowed
                            to move, and the coordinates of the rest will be
                            frozen.""", default=0, valid_values=[0, 1])
        self._cmdValue(key='ntr', datatype='int', realdatatype=TYPE_INT,
                       descr="""Flag for restraining specified atoms in
                             Cartesian space using a harmonic potential,
                             if ntr > 0.""", default=0, valid_range=RANGE_0POS)
        self._cmdValue(key='restraint_wt', datatype='int', realdatatype=TYPE_INT,
                       descr="""The weight (in kcal/mol−A^̊2) for the positional
                             restraints.""", default=None, valid_range=RANGE_0POS)
        self._cmdString(key='restraintmask', datatype='str',
                        descr="""String that specifies the restrained atoms
                              when ntr=1.""", default=None)
        self._cmdString(key='bellymask', datatype='str',
                        descr="""String that specifies the moving atoms when
                        ibelly=1.""", default=None)

        #  Energy minimization (Manual section 18.6.5)
        self._cmdValue(key='maxcyc', datatype='int', realdatatype=TYPE_INT,
                       descr='The maximum number of cycles of minimization.',
                       default=1, valid_range=RANGE_POS)
        self._cmdValue(key='ncyc', datatype='int', realdatatype=TYPE_INT,
                       descr="""If NTMIN is 1 then the method of minimization
                             will be switched from steepest descent to
                             conjugate gradient after NCYC cycles.""",
                       default=10, valid_range=RANGE_POS)
        self._cmdList(key='ntmin', datatype='int',
                      descr='Flag for the method of minimization.',
                      default=1, valid_values=[0, 1, 2, 3, 4])
        self._cmdValue(key='dx0', datatype='float', realdatatype=TYPE_FLOAT,
                       descr='The initial step length.', default=0.01,
                       valid_range=RANGE_POS)
        self._cmdValue(key='drms', datatype='float', realdatatype=TYPE_FLOAT,
                       descr="""The convergence criterion for the energy
                             gradient: minimization will halt when the RMS of
                             the Cartesian elements of the gradient
                             is < DRMS.""", default=1e-4, valid_range=RANGE_POS)

        #  Molecular dynamics (Manual section 18.6.6)
        self._cmdValue(key='nstlim', datatype='int', realdatatype=TYPE_INT,
                       descr='Number of MD-steps to be performed.', default=1,
                       valid_range=RANGE_POS)
        self._cmdValue(key='nscm', datatype='int', realdatatype=TYPE_INT,
                       descr="""Flag for the removal of translational and
                             rotational center-of-mass (COM) motion at regular
                             intervals.""", default=1000, valid_range=RANGE_POS)
        self._cmdValue(key='t', datatype='float', realdatatype=TYPE_FLOAT,
                       descr="""The time at the start (ps) this is for your own
                             reference and is not critical.""", default=0.0,
                       valid_range=RANGE_0POS)
        self._cmdValue(key='dt', datatype='float', realdatatype=TYPE_FLOAT,
                       descr="""The time step (ps).""", default=0.001,
                       valid_range=RANGE_POS)
        self._cmdValue(key='nrespa', datatype='int', realdatatype=TYPE_INT,
                       descr="""This variable allows the user to evaluate
                             slowly-varying terms in the force field less
                             frequently.""", default=None, valid_range=RANGE_POS)

        #  Temperature regulation (Manual section 18.6.7)
        self._cmdList(key='ntt', datatype='int',
                      descr='Switch for temperature scaling.', default=None,
                      valid_values=[0, 1, 2, 3, 9, 10])
        self._cmdValue(key='temp0', datatype='int', realdatatype=TYPE_INT,
                       descr="""Reference temperature at which the system is to
                             be kept, if ntt > 0.""", default=300, valid_range=RANGE_POS)
        self._cmdValue(key='temp0les', datatype='int', realdatatype=TYPE_INT,
                       descr="""This is the target temperature for all LES
                       particles (see Chapter 6 in the Amber Manual).""",
                       default=-1, valid_range=RANGE_ANY)
        self._cmdValue(key='tempi', datatype='float', realdatatype=TYPE_FLOAT,
                       descr="""Initial temperature. For the initial dynamics
                             run, (NTX < 3) the velocities are assigned from
                             a Maxwellian distribution at TEMPI K. If
                             TEMPI = 0.0, the velocities will be calculated
                             from the forces instead. TEMPI has no effect if
                             NTX > 3.""", default=0.0, valid_range=RANGE_0POS)
        self._cmdValue(key='ig', datatype='int', realdatatype=TYPE_INT,
                       descr='The seed for the pseudo-random numbergenerator.',
                       default=-1, valid_range=RANGE_ANY)
        self._cmdValue(key='tautp', datatype='float', realdatatype=TYPE_FLOAT,
                       descr="""Time constant (ps) for heat bath coupling for
                             the system, if ntt = 1.""", default=1.0,
                       valid_range=RANGE_0POS)
        self._cmdValue(key='gamma_ln', datatype='float', realdatatype=TYPE_FLOAT,
                       descr='The collision frequency γ, in ps−1, when ntt = 3.',
                       default=0.0, valid_range=RANGE_0POS)
        self._cmdValue(key='vrand', datatype='int', realdatatype=TYPE_INT,
                       descr="""If vrand>0 and ntt=2, the velocities will be
                             randomized to temperature TEMP0 every vrand steps.""",
                       default=1000, valid_range=RANGE_0POS)
        self._cmdValue(key='vlimit', datatype='float', realdatatype=TYPE_FLOAT,
                       descr="""If not equal to 0.0, then any component of the
                             velocity that is greater than abs(VLIMIT) will be
                             reduced to VLIMIT (preserving the sign).""",
                       default=20.0, valid_range=RANGE_0POS)
        self._cmdValue(key='nkija', datatype='int', realdatatype=TYPE_INT,
                       descr="""For use with ntt=9 and ntt=10., For ntt=9, this
                             is the number of substeps of dt when integrating
                             the thermostat equations of motion, for greater
                             accuracy. For ntt=10, this specifies the number of
                             additional auxiliary velocity variables v1 and v2,
                             which will total nkija×v1 +nkija×v2""",
                       default=1, valid_range=RANGE_0POS)
        self._cmdValue(key='idistr', datatype='int', realdatatype=TYPE_INT,
                       descr="""For use with ntt=9, this is the frequency at
                             which the thermostat velocity distribution
                             functions are accumulated.""",
                       default=None, valid_range=RANGE_0POS)
        self._cmdValue(key='sinrtau', datatype='float', realdatatype=TYPE_FLOAT,
                       descr="""this specifies the time scale for determining
                             the masses associated with the two auxiliary
                             velocity variables v1 and v2 (e.g. thermostat
                             velocities)""", default=1.0, valid_range=RANGE_0POS)

        # Pressure regulation (Manual section 18.6.8)
        self._cmdList(key='ntp', datatype='int',
                      descr='Flag for constant pressure dynamics.',
                      default=0, valid_values=[0, 1, 2, 3])
        self._cmdList(key='barostat', datatype='int',
                      descr="""Flag used to control which barostat to use in
                            order to control the pressure.""", default=1,
                      valid_values=[1, 2])
        self._cmdValue(key='mcbarint', datatype='int', realdatatype=TYPE_INT,
                       descr="""Number of steps between volume change attempts
                             performed as part of the Monte Carlo barostat.""",
                       default=100, valid_range=RANGE_POS)
        self._cmdValue(key='pres0', datatype='float', realdatatype=TYPE_FLOAT,
                       descr="""Reference pressure (in units of bars) at which
                             the system is maintained (when NTP > 0).""",
                       default=1.0, valid_range=RANGE_POS)
        self._cmdValue(key='comp', datatype='float', realdatatype=TYPE_FLOAT,
                       descr="""compressibility of the system when NTP > 0. The
                             units are in 1.0 × 10-6 bar-1""", default=44.6,
                       valid_range=RANGE_0POS)
        self._cmdValue(key='taup', datatype='float', realdatatype=TYPE_FLOAT,
                       descr='Pressure relaxation time (in ps), when NTP > 0.',
                       default=1.0, valid_range=RANGE_POS)
        self._cmdList(key='csurften', datatype='int',
                      descr='Flag for constant surface tension dynamics.',
                      default=0, valid_values=[0, 1, 2, 3])
        self._cmdValue(key='gamma_ten', datatype='float', realdatatype=TYPE_FLOAT,
                       descr='Surface tension value in units of dyne/cm.',
                       default=0.0, valid_range=RANGE_0POS)
        self._cmdValue(key='ninterface', datatype='int', realdatatype=TYPE_INT,
                       descr="""Number of interfaces in the periodic box. There
                             must be at least two interfaces in the periodic box.""",
                       default=2, valid_range=RANGE_0POS)

        # SHAKE bond length constraints (Manual section 18.6.9)
        self._cmdList(key='ntc', datatype='int',
                      descr='Flag for SHAKE to perform bond length constraints.',
                      default=1, valid_values=[1, 2, 3])
        self._cmdValue(key='tol', datatype='float', realdatatype=TYPE_FLOAT,
                       descr="""Relative geometrical tolerance for coordinate
                             resetting in shake.""", default=0.00001,
                       valid_range=RANGE_0POS)
        self._cmdList(key='jfastw', datatype='int',
                      descr='Fast water definition flag.', default=0, valid_values=[0, 4])
        self._cmdString(key='noshakemask', datatype='str',
                        descr="""String that specifies atoms that are not to be
                              shaken (assuming that ntc>1).""", default='')

        # Water cap (Manual section 18.6.10)
        self._cmdList(key='ivcap', datatype='int',
                      descr="""Flag to control cap option. The "cap" refers to
                            a spherical portion of water centered on a point in
                            the solute and restrained by a soft half-harmonic
                            potential. For the best physical realism, this
                            option should be combined with igb=10, in order to
                            include the reaction field of waters that are beyond
                            the cap radius.""", default=0, valid_values=[0, 1, 2, 5])
        self._cmdValue(key='fcap', datatype='float', realdatatype=TYPE_FLOAT,
                       descr='The force constant for the cap restraint potential.',
                       default=None, valid_range=RANGE_POS)
        self._cmdValue(key='cutcap', datatype='float', realdatatype=TYPE_FLOAT,
                       descr='Radius of the cap, if ivcap=1 is used.',
                       default=None, valid_range=RANGE_POS)
        self._cmdValue(key='xcap', datatype='float', realdatatype=TYPE_FLOAT,
                       descr='Location of the cap center, if ivcap=1 is used.',
                       default=None, valid_range=RANGE_ANY)

        # TODO: NMR refinement options (Manual section 18.6.11)
        # TODO: EMAP restraints (Manual section 18.6.12)
        # The pmemd cuda engine does not support these (and many other options
        # that I already have written). Still, users can use sander or the CPU
        # version of pmemd if they like

        # POTENTIAL FUNCTION PARAMETERS (MANUAL SECTION 18.7)
        # Only implementing the bold ones in the Manual (most important ones)
        # for the moment

        # Generic parameters (Manual Section 18.7.1)
        self._cmdList(key='ntf', datatype='int', descr='Force evaluation',
                      default=1, valid_values=[1, 2, 3, 4, 5, 6, 7, 8])
        self._cmdList(key='ntb', datatype='int',
                      descr="""Whether or not periodic boundaries are imposed
                            on the system during the calculation of non-bonded
                            interactions.""", default=None, valid_values=[0, 1, 2])
        self._cmdValue(key='cut', datatype='float', realdatatype=TYPE_FLOAT,
                       descr='Nonbonded cutoff, in Å',
                       default=None, valid_range=RANGE_POS)
        self._cmdList(key='ipol', datatype='int',
                      descr="""When set to 1, use a polarizable force field.
                            See Section 18.7.5 for more information.""",
                      default=0, valid_values=[0, 1])
        self._cmdList(key='ifqnt', datatype='int',
                      descr="""Flag for QM/MM run; if set to 1, you must also
                            include a &qmmm namelist.""", default=0, valid_values=[0, 1])
        self._cmdList(key='igb', datatype='int',
                      descr="""Flag for using the generalized Born or Poisson-Boltzmann
                            implicit solvent models.""",
                      default=0, valid_values=[0, 1, 2, 3, 4, 5, 6, 7, 8, 10])
        self._cmdList(key='irism', datatype='int',
                      descr="""Flag for 3D-reference interaction site model
                            molecular solvation method.""", default=0, valid_values=[0, 1])
        self._cmdList(key='ievb', datatype='int',
                      descr="""If set to 1, use the empirical valence bond method
                            to compute energies and forces.""", default=0, valid_values=[0, 1])
        self._cmdList(key='iamoeba', datatype='int',
                      descr="""Flag for using the amoeba polarizable potentials
                            of Ren and Ponder.""", default=0, valid_values=[0, 1])

        self._cmdString('outputnc', 'str', 'Name of output NetCDF file', None)

        # Assign everything to a single string
        self._cmdString('FORTRAN', 'str', 'Protocol in FORTRAN', None)
        self._cmdString('bash', 'str', '', None)
        # Files
        self._cmdString('bincoordinates', 'str', '', None)  # coordinate binary file .nc (-x)
        self._cmdString('coordinates', 'str', '', None)  # frame coordinates .rst (-r)
        self._cmdString('consref', 'str', '', None)  # constraints reference (-ref)
        self._cmdString('parmfile', 'str', '', None)  # topology file .prmtop (-p)

    def load(self, path='.'):
        """ Loads all files required to run a simulation and apply eventually configured protocols to it

        Parameters
        ----------
        path : str
            Working directory relative to which the configuration file is read
        """
        # load files and reset filenames

        for cmd in self._defaultfnames.keys():
            if cmd in self.__dict__ and self.__dict__[cmd] is not None:
                fname = os.path.join(path, self.__dict__[cmd])
                f = open(fname, 'rb')  # read all as binary
                data = f.read()
                f.close()
                self._files[cmd] = data

                defaultname = self._defaultfnames[cmd]
                if defaultname.endswith('*'):
                    defaultname = '{}.{}'.format(os.path.splitext(defaultname)[0], os.path.splitext(fname)[1][1:])

                self.__dict__[cmd] = defaultname  # use default file names

    def save(self, path, overwrite=False):
        """ Create a directory with all necessary input to run with AMBER.

        Parameters
        ----------
        path : str
            New execution directory to create
        overwrite : bool
            Overwrite output directory if it exists.
        """
        if os.path.exists(path):
            if overwrite:
                shutil.rmtree(path)
            else:
                raise NameError('Directory exists, use overwrite=True or remove directory')
        os.makedirs(path)

        # Write all files using default filenames
        for f in self._files:
            fo = open(os.path.join(path, self.__dict__[f]), 'wb')  # write all as binary
            fo.write(self._files[f])
            fo.close()

        self._files = {}  # empty file dictionary after writing

        self.writeConf(os.path.join(path, 'input'))

    def setup(self, cwd='./', outdir='./run', overwrite=False):
        """ Convenience method performing load and save.

        Parameters
        ----------
        cwd : str
            The directory to load from.
        outdir : str
            The directory to save to.
        overwrite : bool
            Overwrite output directory if it exists.
        """
        self.load(cwd)
        self.save(outdir, overwrite)

    def show(self, quiet=False):
        """ Returns the AMBER configuration file string

        Parameters
        ----------
        quiet : bool
            If true it prints the string to stdout

        Returns
        -------
        conf : str
            The string of the configuration file
        """
        text = ''
        if self.__dict__['FORTRAN'] is not None:
            text = self.__dict__['FORTRAN']

        text += '#\n'

        maxwidth = np.max([len(k) for k in self.__dict__.keys()])

        keys = sorted(list(self.__dict__.keys()))
        # keys = keys + [keys.pop(keys.index('run'))]  # Move the run command to the end
        for cmd in keys:
            if not cmd.startswith('_') and self.__dict__[cmd] is not None and cmd != 'TCL':
                name = cmd
                if cmd == 'scaling14':  # variables cannot start with numbers. We need to rename it here for acemd
                    name = '1-4scaling'
                text += '{name: <{maxwidth}}\t{val:}\n'.format(name=name, val=str(self.__dict__[cmd]),
                                                               maxwidth=maxwidth)

        if not quiet:
            print(text)
        else:
            return text

    def __str__(self):
        return self.show(quiet=True)

    def writeConf(self, fname='input'):
        """ Write an acemd configuration file

        Parameters
        ----------
        fname : output file name
        """
        text = self.show(quiet=True)
        with open(fname, 'w') as f:
            f.write(text)



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import os
import htmd
import shutil
import numpy as np
from htmd.protocols.oldprotocolinterface import ProtocolInterface, TYPE_INT, TYPE_FLOAT, RANGE_0POS, RANGE_POS, RANGE_ANY


class Acemd(ProtocolInterface):
    _defaultfnames = {'bincoordinates': 'input.coor', 'binvelocities': 'input.vel', 'binindex': 'input.idx',
               'structure': 'structure.*', 'parameters': 'parameters', 'extendedsystem': 'input.xsc',
               'coordinates': 'structure.pdb', 'velocities': 'velocity.pdb', 'consref': 'structure.pdb',
               'parmfile': 'structure.prmtop'}

    def __init__(self, version=2):
        super().__init__()
        self._version = version
        self._files = {}
        self._outnames = {}

        # Options
        self._cmdString('temperature', 'float', 'Temperature of the thermostat in Kelvin.', None)
        self._cmdString('restart', 'str', 'Restart simulation.', None)
        self._cmdString('restartfreq', 'str', 'Restart file frequency.', None)
        self._cmdString('outputname', 'str', 'Output file name.', None)
        self._cmdString('xtcfile', 'str', 'Output XTC file name.', None)
        self._cmdString('xtcfreq', 'str', 'XTC sampling frequency in steps.', None)
        self._cmdString('timestep', 'str', 'Simulation timestep.', None)
        self._cmdString('rigidbonds', 'str', '', None)
        self._cmdString('hydrogenscale', 'str', '', None)
        self._cmdString('switching', 'str', '', None)
        self._cmdString('switchdist', 'str', '', None)
        self._cmdString('cutoff', 'str', '', None)
        self._cmdString('exclude', 'str', '', None)
        self._cmdString('scaling14', 'str', '', None)
        self._cmdString('langevin', 'str', '', None)
        self._cmdString('langevintemp', 'str', '', None)
        self._cmdString('langevindamping', 'str', '', None)
        self._cmdString('pme', 'str', '', None)
        self._cmdString('pmegridspacing', 'str', '', None)
        self._cmdString('fullelectfrequency', 'str', '', None)
        self._cmdString('energyfreq', 'str', '', None)
        if self._version == 2:
            self._cmdString('constraints', 'str', '', None)
            self._cmdString('consref', 'str', '', None)
            self._cmdString('constraintscaling', 'str', '', None)
        if self._version == 3:
            self._cmdString('atomrestraint', 'str', '', None)
            self._cmdString('grouprestraint', 'str', '', None)
        self._cmdString('berendsenpressure', 'str', '', None)
        self._cmdString('berendsenpressuretarget', 'str', '', None)
        self._cmdString('berendsenpressurerelaxationtime', 'str', '', None)
        self._cmdString('tclforces', 'str', '', None)
        self._cmdString('minimize', 'str', '', None)
        self._cmdString('run', 'str', '', None)
        self._cmdString('celldimension', 'str', '', None)
        self._cmdString('useconstantratio', 'str', '', None)
        self._cmdString('amber', 'str', '', None)
        self._cmdString('dielectric', 'str', '', None)
        self._cmdString('pairlistdist', 'str', '', None)
        if self._version == 2:
            self._cmdString('TCL', 'str', '', None)

        # Files
        self._cmdString('bincoordinates', 'str', '', None)
        self._cmdString('binvelocities', 'str', '', None)
        self._cmdString('binindex', 'str', '', None)
        self._cmdString('structure', 'str', '', None)
        self._cmdString('parameters', 'str', '', None)
        self._cmdString('extendedsystem', 'str', '', None)
        self._cmdString('coordinates', 'str', '', None)
        self._cmdString('velocities', 'str', '', None)
        self._cmdString('parmfile', 'str', '', None)

    def load(self, path='.'):
        """ Loads all files required to run a simulation and apply eventually configured protocols to it

        Parameters
        ----------
        path : str
            Working directory relative to which the configuration file is read
        """
        # load files and reset filenames
        for cmd in self._defaultfnames.keys():
            if cmd in self.__dict__ and self.__dict__[cmd] is not None:
                fname = os.path.join(path, self.__dict__[cmd])
                f = open(fname, 'rb')  # read all as binary
                data = f.read()
                f.close()
                self._files[cmd] = data

                defaultname = self._defaultfnames[cmd]
                if defaultname.endswith('*'):
                    defaultname = '{}.{}'.format(os.path.splitext(defaultname)[0], os.path.splitext(fname)[1][1:])
                self._outnames[cmd] = defaultname

    def save(self, path, overwrite=False):
        """ Create a directory with all necessary input to run acemd.
    
        Parameters
        ----------
        path : str
            New execution directory to create
        overwrite : bool
            Overwrite output directory if it exists.
        """
        if os.path.exists(path):
            if overwrite:
                shutil.rmtree(path)
            else:
                raise NameError('Directory exists, use overwrite=True or remove directory')
        os.makedirs(path)

        # Write all files using default filenames
        for f in self._files:
            fo = open(os.path.join(path, self._outnames[f]), 'wb')  # write all as binary
            fo.write(self._files[f])
            fo.close()

        self._writeBashRun(os.path.join(path, 'run.sh'))

        self._files = {}  # empty file dictionary after writing
    
        self.writeConf(os.path.join(path, 'input'))

    def setup(self, cwd='./', outdir='./run', overwrite=False):
        """ Convenience method performing load and save.

        Parameters
        ----------
        cwd : str
            The directory to load from.
        outdir : str
            The directory to save to.
        overwrite : bool
            Overwrite output directory if it exists.
        """
        self.load(cwd)
        self.save(outdir, overwrite)

    def show(self, quiet=False):
        """ Returns the Acemd configuration file string

        Parameters
        ----------
        quiet : bool
            If true it prints the string to stdout

        Returns
        -------
        conf : str
            The string of the configuration file
        """
        text = ''
        if self.__dict__['TCL'] is not None:
            text = self.__dict__['TCL']

        text += '#\n'

        maxwidth = np.max([len(k) for k in self.__dict__.keys()])

        keys = sorted(list(self.__dict__.keys()))
        keys = keys + [keys.pop(keys.index('run'))]  # Move the run command to the end
        for cmd in keys:
            if not cmd.startswith('_') and self.__dict__[cmd] is not None and cmd != 'TCL':
                val = self.__dict__[cmd]
                if cmd in self._outnames:
                    val = self._outnames[cmd]
                name = cmd
                if cmd == 'scaling14':  # variables cannot start with numbers. We need to rename it here for acemd
                    name = '1-4scaling'
                text += '{name: <{maxwidth}}\t{val:}\n'.format(name=name, val=val, maxwidth=maxwidth)

        if not quiet:
            print(text)
        else:
            return text

    def _writeBashRun(self, fname):
        with open(fname, 'w') as f:
            if self._version == 3:
                f.write('#!/bin/bash\nacemd3 >log.txt 2>&1')
            else:
                f.write('#!/bin/bash\nacemd >log.txt 2>&1')
        os.chmod(fname, 0o700)

    def __repr__(self):
        return self.show(quiet=True)

    def writeConf(self, fname='input'):
        """ Write an acemd configuration file

        Parameters
        ----------
        fname : output file name
        """
        text = self.show(quiet=True)
        fo = open(fname, 'w')
        fo.write(text)
        fo.close()

if __name__ == "__main__":
    #l=Acemd.protocols(quiet=True)
    import htmd.home
    acemd = Acemd()
    acemd.structure = '5dhfr_cube.psf'
    acemd.parameters = 'par_all22_prot.inp'
    homedir = htmd.home.home()
    acemd.coordinates = '5dhfr_cube.pdb'
    acemd.setup(homedir + '/data/dhfr', '/tmp/testdir', overwrite=True)
    print(acemd)

